""""""
from __future__ import annotations

from typing import Callable

import numpy as np

from gpaw.core import UGArray, UGDesc
from gpaw.gpu import einsum
from gpaw.new import zips
from gpaw.new.c import (add_to_density, add_to_density_gpu, evaluate_lda_gpu,
                        evaluate_pbe_gpu)
from gpaw.new.calculation import DFTState
from gpaw.new.ibzwfs import IBZWaveFunctions
from gpaw.new.pwfd.wave_functions import PWFDWaveFunctions
from gpaw.typing import Array2D
from gpaw.xc import XC
from gpaw.xc.functional import XCFunctional as OldXCFunctional
from gpaw.xc.gga import add_gradient_correction
from gpaw.xc.mgga import MGGA
from gpaw.xc.vdw import VDWFunctionalBase
from gpaw.hybrids import HybridXC


def create_functional(xc: OldXCFunctional | str | dict,
                      grid: UGDesc,
                      xp=np) -> Functional:
    exx_fraction = 0.0
    exx_omega = 0.0
    if isinstance(xc, (str, dict)):
        xc = XC(xc, xp=xp)

    if xc.type == 'HYB':
        assert isinstance(xc, HybridXC)
        exx_fraction = xc.exx_fraction
        exx_omega = xc.omega
        xc = xc.xc

    if xc.type == 'LDA':
        functional = LDAFunctional(xc, grid, xp)
    elif xc.type == 'GGA':
        functional = GGAFunctional(xc, grid, xp)
    elif xc.type == 'MGGA':
        functional = MGGAFunctional(xc, grid)
    else:
        raise ValueError(f'{xc.type} not supported')

    functional.exx_fraction = exx_fraction
    functional.exx_omega = exx_omega

    return functional


class Functional:
    def __init__(self,
                 xc: OldXCFunctional,
                 grid: UGDesc,
                 xp=np):
        self.xc = xc
        self.grid = grid
        self.xp = xp
        self.setup_name = self.xc.get_setup_name()
        self.name = self.xc.name
        self.type = self.xc.type
        self.xc.set_grid_descriptor(grid._gd)
        self.exx_fraction = 0.0
        self.exx_omega = 0.0

    def __str__(self):
        return f'name: {self.xc.get_description()}'

    def calculate_paw_correction(self, setup, d, h=None):
        return self.xc.calculate_paw_correction(setup, d, h)

    def get_setup_name(self) -> str:
        return self.name

    def stress_contribution(self,
                            state: DFTState,
                            interpolate: Callable[[UGArray], UGArray]
                            ) -> Array2D:
        args, kwargs = self._args(state, interpolate)
        if state.ibzwfs.kpt_band_comm.rank == 0:
            if self.xp is np:
                self.xc.kernel.calculate(*[array.data for array in args])
            # Special GPU cases:
            elif self.name == 'LDA':
                e_r, nt_sr, vt_sr = args
                evaluate_lda_gpu(nt_sr.data, vt_sr.data, e_r.data)
            elif self.name == 'PBE':
                e_r, nt_sr, vt_sr, sigma_xr, dedsigma_xr = args
                evaluate_pbe_gpu(nt_sr.data, vt_sr.data, e_r.data,
                                 sigma_xr.data, dedsigma_xr.data)
            else:
                1 / 0
            return self._stress(*args, **kwargs)
        return self.xp.zeros((3, 3))

    def _args(self,
              state: DFTState,
              interpolate: Callable[[UGArray], UGArray]
              ) -> tuple[tuple[UGArray, ...], dict]:
        raise NotImplementedError

    def _stress(self, *args: UGArray, **kwargs) -> Array2D:
        raise NotImplementedError


class LDAFunctional(Functional):
    def calculate(self,
                  nt_sr: UGArray,
                  taut_sr: UGArray | None = None) -> tuple[float,
                                                           UGArray,
                                                           UGArray | None]:
        vxct_sr = nt_sr.new()
        vxct_sr.data[:] = 0.0
        e_r = nt_sr.desc.empty(xp=self.xp)
        if self.xp is np:
            self.xc.kernel.calculate(e_r.data, nt_sr.data, vxct_sr.data)
        else:
            if self.name != 'LDA':
                raise ValueError(f'{self.name} not supported on GPU')
            evaluate_lda_gpu(nt_sr.data, vxct_sr.data, e_r.data)
        exc = e_r.integrate()
        return exc, vxct_sr, None

    def _args(self, state, interpolate):
        ibzwfs = state.ibzwfs
        if ibzwfs.kpt_band_comm.rank != 0:
            return (), {}
        nt_sR = state.density.nt_sR
        e_r = self.grid.empty(xp=self.xp)
        nt_sr = interpolate(nt_sR)
        vt_sr = nt_sr.new(zeroed=True)
        return (e_r, nt_sr, vt_sr), {}

    def _stress(self,  # type: ignore
                e_r: UGArray,
                nt_sr: UGArray,
                vt_sr: UGArray) -> Array2D:
        P = e_r.integrate(skip_sum=True)
        for vt_r, nt_r in zip(vt_sr, nt_sr):
            P -= vt_r.integrate(nt_r, skip_sum=True)
        return float(P) * self.xp.eye(3)


class GGAFunctional(LDAFunctional):
    def __init__(self,
                 xc: OldXCFunctional,
                 grid: UGDesc,
                 xp=np):
        super().__init__(xc, grid, xp)
        # xc already has Gradient.apply bound methods!!!
        self.grad_v = [grad.__self__ for grad in xc.grad_v]  # type: ignore

    def calculate(self,
                  nt_sr: UGArray,
                  taut_sr: UGArray | None = None) -> tuple[float,
                                                           UGArray,
                                                           UGArray | None]:
        gradn_svr, sigma_xr = gradient_and_sigma(self.grad_v, nt_sr)

        vxct_sr = nt_sr.new(zeroed=True)
        dedsigma_xr = sigma_xr.new()
        e_r = self.grid.empty(xp=self.xp)

        if self.xp is np:
            args = [a.data
                    for a in [e_r, nt_sr, vxct_sr, sigma_xr, dedsigma_xr]]
            if 'vdW' not in self.name:
                self.xc.kernel.calculate(*args)
            else:
                assert isinstance(self.xc, VDWFunctionalBase)
                self.xc.calculate_exchange(*args)
                self.xc.calculate_correlation(*args)
        else:
            if self.name != 'PBE':
                raise ValueError(f'{self.name} not supported on GPU')
            evaluate_pbe_gpu(nt_sr.data, vxct_sr.data, e_r.data,
                             sigma_xr.data, dedsigma_xr.data)

        add_gradient_correction([grad.apply for grad in self.grad_v],
                                gradn_svr.data, sigma_xr.data,
                                dedsigma_xr.data, vxct_sr.data)
        exc = e_r.integrate()
        return exc, vxct_sr, None

    def _args(self,
              state: DFTState,
              interpolate: Callable[[UGArray], UGArray]
              ) -> tuple[tuple[UGArray, ...], dict]:
        args, kwargs = LDAFunctional._args(self, state, interpolate)
        if args:
            e_r, nt_sr, vt_sr = args
            gradn_svr, sigma_xr = gradient_and_sigma(self.grad_v, nt_sr)
            dedsigma_xr = sigma_xr.new()
            args += (sigma_xr, dedsigma_xr)
            kwargs = {'gradn_svr': gradn_svr}
        return args, kwargs

    def _stress(self,  # type: ignore
                e_r, nt_sr, vt_sr, sigma_xr, dedsigma_xr,
                gradn_svr,
                ) -> Array2D:
        stress_vv = LDAFunctional._stress(self, e_r, nt_sr, vt_sr)
        P = 0.0
        for sigma_r, dedsigma_r in zip(sigma_xr, dedsigma_xr):
            P -= 2 * sigma_r.integrate(dedsigma_r, skip_sum=True)
        stress_vv += float(P) * self.xp.eye(3)

        nspins = len(nt_sr)
        for v1 in range(3):
            for v2 in range(3):
                stress_vv[v1, v2] -= 2 * dedsigma_xr[0].integrate(
                    gradn_svr[0, v1] * gradn_svr[0, v2], skip_sum=True)
                if nspins == 2:
                    stress_vv[v1, v2] -= 2 * dedsigma_xr[1].integrate(
                        gradn_svr[0, v1] * gradn_svr[1, v2], skip_sum=True)
                    stress_vv[v1, v2] -= 2 * dedsigma_xr[2].integrate(
                        gradn_svr[1, v1] * gradn_svr[1, v2], skip_sum=True)

        return stress_vv


def gradient_and_sigma(grad_v, n_sr: UGArray) -> tuple[UGArray, UGArray]:
    """Calculate gradient of density and sigma.

    Returns:::

      _    _
      ∇ n (r)
         s

    and:::

         _     _   2     _    _          _   2
      σ (r) = |∇n |, σ = ∇n . ∇n ,  σ = |∇n |
       0         0    1    0    1    2     1
    """
    nspins = n_sr.dims[0]
    xp = n_sr.xp

    gradn_svr = n_sr.desc.empty((nspins, 3), xp=xp)
    for v, grad in enumerate(grad_v):
        for s in range(nspins):
            grad(n_sr[s], gradn_svr[s, v])

    sigma_xr = n_sr.desc.empty(nspins * 2 - 1, xp=xp)
    dot_product(gradn_svr[0], None, sigma_xr[0])
    if nspins == 2:
        dot_product(gradn_svr[0], gradn_svr[1], sigma_xr[1])
        dot_product(gradn_svr[1], None, sigma_xr[2])

    return gradn_svr, sigma_xr


def dot_product(a_vr, b_vr, out_r):
    xp = a_vr.xp
    if b_vr is None:
        out_r.data[:] = 0.0
        if xp is np:
            for a_r in a_vr.data:
                add_to_density(1.0, a_r, out_r.data)
        else:
            add_to_density_gpu(xp.ones(3), a_vr.data, out_r.data)
    else:
        einsum('vabc, vabc -> abc', a_vr.data, b_vr.data, out=out_r.data)


class MGGAFunctional(GGAFunctional):
    def get_setup_name(self):
        return 'PBE'

    def calculate(self,
                  nt_sr: UGArray,
                  taut_sr: UGArray | None = None) -> tuple[float,
                                                           UGArray,
                                                           UGArray | None]:
        gradn_svr, sigma_xr = gradient_and_sigma(self.grad_v, nt_sr)
        assert isinstance(self.xc, MGGA), self.xc
        e_r = self.grid.empty()
        if taut_sr is None:
            taut_sr = nt_sr.new(zeroed=True)
        dedtaut_sr = taut_sr.new()
        vxct_sr = taut_sr.new()
        vxct_sr.data[:] = 0.0
        dedsigma_xr = sigma_xr.new()
        self.xc.kernel.calculate(e_r.data, nt_sr.data, vxct_sr.data,
                                 sigma_xr.data, dedsigma_xr.data,
                                 taut_sr.data, dedtaut_sr.data)
        add_gradient_correction([grad.apply for grad in self.grad_v],
                                gradn_svr.data, sigma_xr.data,
                                dedsigma_xr.data, vxct_sr.data)
        return e_r.integrate(), vxct_sr, dedtaut_sr

    def _args(self, state: DFTState, interpolate: Callable[[UGArray], UGArray]
              ):
        args, kwargs = GGAFunctional._args(self, state, interpolate)
        taut_swR = _taut(state.ibzwfs, state.density.nt_sR.desc)

        if not args:
            return (), {}

        e_r, nt_sr, vt_sr, sigma_xr, dedsigma_xr = args
        taut_sR = state.density.taut_sR
        assert taut_sR is not None
        taut_sr = interpolate(taut_sR)
        dedtaut_sr = taut_sr.new()
        args += (taut_sr, dedtaut_sr)
        kwargs['taut_swR'] = taut_swR
        kwargs['interpolate'] = interpolate

        return args, kwargs

    def _stress(self,  # type: ignore
                e_r,
                nt_sr, vt_sr,
                sigma_xr, dedsigma_xr,
                taut_sr, dedtaut_sr,
                gradn_svr,
                taut_swR,
                interpolate
                ) -> Array2D:  # type: ignore
        stress_vv = GGAFunctional._stress(
            self, e_r, nt_sr, vt_sr, sigma_xr, dedsigma_xr,
            gradn_svr=gradn_svr)
        for taut_wR, dedtaut_r in zips(taut_swR, dedtaut_sr):
            w = 0
            for v1 in range(3):
                for v2 in range(v1, 3):
                    taut_r = interpolate(taut_wR[w])
                    s = taut_r.integrate(dedtaut_r, skip_sum=True)
                    stress_vv[v1, v2] -= s
                    if v2 != v1:
                        stress_vv[v2, v1] -= s
                    w += 1

        P = 0.0
        for taut_r, dedtaut_r in zip(taut_sr, dedtaut_sr):
            P -= taut_r.integrate(dedtaut_r, skip_sum=True)
        stress_vv += P * np.eye(3)

        return stress_vv


def _taut(ibzwfs: IBZWaveFunctions, grid: UGDesc) -> UGArray | None:
    """Calculate upper half of symmetric ked tensor.

    Returns ``taut_swR=taut_svvR``.  Mapping from ``w`` to ``vv``::

        0 1 2
        . 3 4
        . . 5

    Only cores with ``kpt_comm.rank==0`` and ``band_comm.rank==0``
    return the uniform grids - other cores return None.
    """
    # "1" refers to undistributed arrays
    dpsit1_vR = grid.new(comm=None, dtype=ibzwfs.dtype).empty(3)
    taut1_swR = grid.new(comm=None).zeros((ibzwfs.nspins, 6))
    assert isinstance(taut1_swR, UGArray)  # Argggghhh!
    domain_comm = grid.comm

    for wfs in ibzwfs:
        assert isinstance(wfs, PWFDWaveFunctions)
        psit_nG = wfs.psit_nX
        pw = psit_nG.desc

        pw1 = pw.new(comm=None)
        psit1_G = pw1.empty()
        iGpsit1_G = pw1.empty()
        Gplusk1_Gv = pw1.reciprocal_vectors()

        occ_n = wfs.weight * wfs.spin_degeneracy * wfs.myocc_n

        (N,) = psit_nG.mydims
        for n1 in range(0, N, domain_comm.size):
            n2 = min(n1 + domain_comm.size, N)
            psit_nG[n1:n2].gather_all(psit1_G)
            n = n1 + domain_comm.rank
            if n >= N:
                continue
            f = occ_n[n]
            if f == 0.0:
                continue
            for Gplusk1_G, dpsit1_R in zips(Gplusk1_Gv.T, dpsit1_vR):
                iGpsit1_G.data[:] = psit1_G.data
                iGpsit1_G.data *= 1j * Gplusk1_G
                iGpsit1_G.ifft(out=dpsit1_R)
            w = 0
            for v1 in range(3):
                for v2 in range(v1, 3):
                    taut1_swR[wfs.spin, w].data += (
                        f * (dpsit1_vR[v1].data.conj() *
                             dpsit1_vR[v2].data).real)
                    w += 1

    ibzwfs.kpt_comm.sum(taut1_swR.data, 0)
    if ibzwfs.kpt_comm.rank == 0:
        ibzwfs.band_comm.sum(taut1_swR.data, 0)
        if ibzwfs.band_comm.rank == 0:
            domain_comm.sum(taut1_swR.data, 0)
            if domain_comm.rank == 0:
                symmetries = ibzwfs.ibz.symmetries
                taut1_swR.symmetrize(symmetries.rotation_scc,
                                     symmetries.translation_sc)
            taut_swR = grid.empty((ibzwfs.nspins, 6))
            taut_swR.scatter_from(taut1_swR)
            return taut_swR
    return None



import pytest
from ase.build import molecule

from gpaw import GPAW, Davidson, MixerSum

# Move atom infinitesimally across cell border and test that SCF loop is still
# well converged afterwards.  If it is /not/ well converged, then the code
# which compensates for discontinuity of phases is probably broken.


# @pytest.mark.later
@pytest.mark.parametrize(
    'params',
    [dict(mode='pw', eigensolver=Davidson(3),
          experimental={'reuse_wfs_method': 'paw'}),
     # pw + lcao extrapolation is currently broken (PWLFC lacks integrate2):
     # dict(mode='pw', experimental={'reuse_wfs_method': 'lcao'}),
     dict(mode='fd', h=0.3,
          experimental={'reuse_wfs_method': 'lcao'}),
     dict(mode='lcao', basis='sz(dzp)', h=0.3)])
def test_generic_move_across_cell(gpaw_new, params):
    params.update(
        xc='oldLDA',
        # make sure MixerSum works for spin-paired system also:
        mixer=MixerSum(0.7),
        kpts=[1, 1, 2])
    calc = GPAW(**params)
    atoms = molecule('H2O', vacuum=2.5)
    atoms.pbc = 1

    # Translate O to corner:
    atoms.positions -= atoms.positions[0, None, :]

    # Be sure that we are on the positive axis:

    atoms.calc = calc

    eps = 1e-12
    atoms.positions[0, :] = eps
    atoms.get_potential_energy()
    atoms.positions[0, 2] -= 2 * eps
    atoms.get_potential_energy()

    # We should be within the convergence criterion.
    # It runs a minimum of three iterations:
    if gpaw_new:
        assert calc.dft.scf_loop.niter == 3
    else:
        assert calc.scf.niter == 3



from contextlib import contextmanager
from functools import wraps
from io import StringIO
from typing import Callable, TypeVar


class GlobalTimer:
    def __init__(self):
        from gpaw.utilities.timing import Timer
        self._timers = [Timer()]

    @contextmanager
    def context(self, timer):
        # XXX We need to decide what "timer contexts" are,
        # at least that will be necessary if we want a default
        # behaviour which would then be in effect during runs of the
        # test suite.
        self._timers.append(timer)
        try:
            yield
        finally:
            self._timers.pop()

    def start(self, name):
        self._timers[-1].start(name)

    def stop(self, name=None):
        self._timers[-1].stop(name=name)

    def tostring(self):
        buf = StringIO()
        self._timers[-1].write(out=buf)
        return buf.getvalue()


T = TypeVar('T')


def trace(meth: Callable[..., T]) -> Callable[..., T]:
    """Decorator for telling global timer to trace a function or method."""

    modname = meth.__module__
    methname = meth.__qualname__
    name = f'{modname}.{methname}'

    @wraps(meth)
    def wrapper(*args, **kwargs):
        global_timer.start(name)
        try:
            return meth(*args, **kwargs)
        finally:
            global_timer.stop()

    return wrapper


global_timer = GlobalTimer()



from __future__ import annotations

from gpaw.core import UGArray
from gpaw.core.arrays import DistributedArrays as XArray


class Hamiltonian:
    def apply(self,
              vt_sR: UGArray,
              dedtaut_sR: UGArray | None,
              ibzwfs,
              D_asii,
              psit_nG: XArray,
              out: XArray,
              spin: int) -> XArray:
        self.apply_local_potential(vt_sR[spin], psit_nG, out)
        if dedtaut_sR is not None:
            self.apply_mgga(dedtaut_sR[spin], psit_nG, out)
        self.apply_orbital_dependent(ibzwfs, D_asii, psit_nG, spin, out)
        return out

    def apply_local_potential(self,
                              vt_R: UGArray,
                              psit_nG: XArray,
                              out: XArray) -> None:
        raise NotImplementedError

    def apply_mgga(self,
                   dedtaut_R: UGArray,
                   psit_nG: XArray,
                   vt_nG: XArray) -> None:
        raise NotImplementedError

    def apply_orbital_dependent(self,
                                ibzwfs,
                                D_asii,
                                psit_nG: XArray,
                                spin: int,
                                out: XArray) -> None:
        pass

    def create_preconditioner(self, blocksize):
        raise NotImplementedError



import numpy as np
import numpy.linalg as la
from ase.units import Hartree

from gpaw.lcao.overlap import NewTwoCenterIntegrals as TwoCenterIntegrals
from gpaw.lfc import BasisFunctions
from gpaw.utilities import unpack_hermitian
from gpaw.utilities.tools import dagger, lowdin, tri2full
from gpaw.lcao.tools import basis_subset2


def dots(*args):
    x = args[0]
    for M in args[1:]:
        x = np.dot(x, M)
    return x


def normalize(U, U2=None, norms=None):
    if norms is None:
        norms2 = np.ones(U.shape[1])
    else:
        norms2 = np.sqrt(norms)

    if U2 is None:
        for n, col in zip(norms2, U.T):
            col *= n / la.norm(col)
    else:
        for n, col1, col2 in zip(norms2, U.T, U2.T):
            norm = np.sqrt(np.vdot(col1, col1) + np.vdot(col2, col2))
            col1 *= n / norm
            col2 *= n / norm


def normalize2(C, S):
    C /= np.sqrt(dots(dagger(C), S, C).diagonal())


def get_rot(F_MM, V_oM, L):
    eps_M, U_MM = la.eigh(F_MM)
    indices = eps_M.real.argsort()[-L:]
    U_Ml = U_MM[:, indices]
    U_Ml /= np.sqrt(dots(U_Ml.T.conj(), F_MM, U_Ml).diagonal())

    U_ow = V_oM.copy()
    U_lw = np.dot(U_Ml.T.conj(), F_MM)
    for col1, col2 in zip(U_ow.T, U_lw.T):
        norm = np.sqrt(np.vdot(col1, col1) + np.vdot(col2, col2))
        col1 /= norm
        col2 /= norm
    return U_ow, U_lw, U_Ml


def condition_number(S):
    eps = la.eigvalsh(S).real
    return eps.max() / eps.min()


def eigvals(H, S):
    return np.sort(la.eigvals(la.solve(S, H)).real)


def get_bfs(calc):
    wfs = calc.wfs
    bfs = BasisFunctions(wfs.gd,
                         [setup.basis_functions_J for setup in wfs.setups],
                         wfs.kd, cut=True)
    bfs.set_positions(wfs.spos_ac)
    return bfs


def get_lcao_projections_HSP(calc, bfs=None, spin=0, projectionsonly=True):
    """Some title.

    if projectionsonly is True, return the projections::

      V_qnM = <psi_qn | Phi_qM>

    else, also return the Hamiltonian, overlap, and projector overlaps::

      H_qMM  = <Phi_qM| H |Phi_qM'>
      S_qMM  = <Phi_qM|Phi_qM'>
      P_aqMi = <pt^a_qi|Phi_qM>
    """
    if calc.wfs.kd.comm.size != 1:
        raise NotImplementedError('Parallelization over spin/kpt not '
                                  'implemented yet.')
    comm = calc.wfs.gd.comm
    nq = len(calc.wfs.kd.ibzk_qc)
    nao = calc.wfs.setups.nao
    dtype = calc.wfs.dtype
    if bfs is None:
        bfs = get_bfs(calc)
    tci = TwoCenterIntegrals(calc.wfs.gd.cell_cv,
                             calc.wfs.gd.pbc_c,
                             calc.wfs.setups,
                             calc.wfs.kd.ibzk_qc,
                             calc.wfs.kd.gamma)

    # Calculate projector overlaps, and (lower triangle of-) S and T matrices
    S_qMM = np.zeros((nq, nao, nao), dtype)
    T_qMM = np.zeros((nq, nao, nao), dtype)
    P_aqMi = {}

    for a in bfs.my_atom_indices:
        ni = calc.wfs.setups[a].ni
        P_aqMi[a] = np.zeros((nq, nao, ni), dtype)
    tci.calculate(calc.spos_ac, S_qMM, T_qMM, P_aqMi)

    from gpaw.lcao.atomic_correction import DenseAtomicCorrection
    dS_aii = {a: calc.wfs.setups[a].dO_ii for a in bfs.my_atom_indices}
    atomic_correction = DenseAtomicCorrection(P_aqMi, dS_aii, 0, nao)
    atomic_correction.add_overlap_correction(S_qMM)
    # calc.wfs.P_aqMi = P_aqMi
    # for q, S_MM in enumerate(S_qMM):
    #     atomic_correction.calculate_manual(wfs, q,
    # atomic_correction.add_overlap_correction(calc.wfs, S_qMM)
    calc.wfs.gd.comm.sum(S_qMM)
    calc.wfs.gd.comm.sum(T_qMM)

    # Calculate projections
    V_qnM = np.zeros((nq, calc.wfs.bd.nbands, nao), dtype)
    for kpt in calc.wfs.kpt_u:
        if kpt.s != spin:
            continue
        V_nM = V_qnM[kpt.q]
        # bfs.integrate2(kpt.psit_nG[:], V_nM, kpt.q) # all bands to save time
        for n, V_M in enumerate(V_nM):  # band-by-band to save memory
            bfs.integrate2(kpt.psit_nG[n][:], V_M, kpt.q)
        for a, P_ni in kpt.projections.items():
            dS_ii = calc.wfs.setups[a].dO_ii
            P_Mi = P_aqMi[a][kpt.q]
            V_nM += np.dot(P_ni, np.inner(dS_ii, P_Mi).conj())
    comm.sum(V_qnM)
    if projectionsonly:
        return V_qnM

    # Determine potential matrix
    vt_G = calc.hamiltonian.vt_sG[spin]
    H_qMM = np.zeros((nq, nao, nao), dtype)
    for q, H_MM in enumerate(H_qMM):
        bfs.calculate_potential_matrix(vt_G, H_MM, q)

    # Make Hamiltonian as sum of kinetic (T) and potential (V) matrices
    # and add atomic corrections
    for a, P_qMi in P_aqMi.items():
        dH_ii = unpack_hermitian(calc.hamiltonian.dH_asp[a][spin])
        for P_Mi, H_MM in zip(P_qMi, H_qMM):
            H_MM += np.dot(P_Mi, np.inner(dH_ii, P_Mi).conj())
    comm.sum(H_qMM)
    H_qMM += T_qMM

    # Fill in the upper triangles of H and S
    for H_MM, S_MM in zip(H_qMM, S_qMM):
        tri2full(H_MM)
        tri2full(S_MM)
    H_qMM *= Hartree

    return V_qnM, H_qMM, S_qMM, P_aqMi


def convert_projection_data(symbols, V_qnM, H_qMM, S_qMM, P_aqMi,
                            originaltype='dzp', newtype='sz'):
    Nq = len(H_qMM)
    mask = basis_subset2(symbols, originaltype, newtype)
    mask2 = np.outer(mask, mask)
    NMnew = len(mask.nonzero()[0])
    V_qnM = V_qnM[..., mask]
    H_qMM = H_qMM[:, mask2].reshape(Nq, NMnew, NMnew)
    S_qMM = S_qMM[:, mask2].reshape(Nq, NMnew, NMnew)
    P2_aqMi = {}
    for a, P_qMi in P_aqMi.items():
        P2_aqMi[a] = P_qMi[:, mask, :]
    return V_qnM, H_qMM, S_qMM, P2_aqMi


def get_phs(calc, s=0):
    return get_lcao_projections_HSP(calc, bfs=None,
                                    spin=s, projectionsonly=False)


class ProjectedWannierFunctions:
    def __init__(self, projections, h_lcao, s_lcao, eigenvalues, kpoints,
                 L_k=None, M_k=None, N=None, fixedenergy=None):
        """projections[n,i] = <psi_n|f_i>
           h_lcao[i1, i2] = <f_i1|h|f_i2>
           s_lcao[[i1, i2] = <f_i1|f_i2>
           eps_n: Exact eigenvalues
           L: Number of extra degrees of freedom
           M: Number of states to exactly span
           N: Total number of bands in the calculation

           Methods:
           -- get_hamiltonian_and_overlap_matrix --
           will return the hamiltonian and identity operator
           in the projected wannier function basis.
           The following steps are performed:

           1) calculate_edf       -> self.b_il
           2) calculate_rotations -> self.Uo_mi and self.Uu_li
           3) calculate_overlaps  -> self.S_ii
           4) calculate_hamiltonian_matrix -> self.H_ii

           -- get_eigenvalues --
           gives the eigenvalues of of the hamiltonian in the
           projected wannier function basis.

           -- indices --
           i localized function index
           n eigenstate index
           l edf index
           m fixed eigenstate index
           k k-point index
           """

        self.eps_kn = eigenvalues
        self.ibzk_kc = kpoints
        self.nk = len(self.ibzk_kc)
        self.V_kni = projections  # <psi_n1|f_i1>
        self.dtype = self.V_kni.dtype
        self.Nw = self.V_kni.shape[2]
        self.s_lcao_kii = s_lcao  # F_ii[i1,i2] = <f_i1|f_i2>
        self.h_lcao_kii = h_lcao

        if N is None:
            N = self.V_kni.shape[1]
        self.N = N

        if fixedenergy is None:
            raise NotImplementedError(
                'Only fixedenergy is implemented for now')
        else:
            self.fixedenergy = fixedenergy
            self.M_k = [sum(eps_n <= fixedenergy) for eps_n in self.eps_kn]
            self.L_k = [self.Nw - M for M in self.M_k]
            print("fixedenergy =", self.fixedenergy)

        print('N =', self.N)
        print('skpt_kc = ')
        print(self.ibzk_kc)
        print('M_k =', self.M_k)
        print('L_k =', self.L_k)
        print('Nw =', self.Nw)

    def get_hamiltonian_and_overlap_matrix(self, useibl=True):
        self.calculate_edf(useibl=useibl)
        self.calculate_rotations()
        self.calculate_overlaps()
        self.calculate_hamiltonian_matrix(useibl=useibl)
        return self.H_kii, self.S_kii

    def calculate_edf(self, useibl=True):
        """Calculate the coefficients b_il in the expansion of the EDF.

        ``|phi_l> = sum_i b_il |f^u_i>``, in terms of ``|f^u_i> = P^u|f_i>``.

        To use the infinite band limit set useibl=True.
        N is the total number of bands to use.
        """

        for k, L in enumerate(self.L_k):
            if L == 0:
                assert L != 0, 'L_k=0 for k=%i. Not implemented' % k

        self.Vo_kni = [V_ni[:M] for V_ni, M in zip(self.V_kni, self.M_k)]

        self.Fo_kii = np.asarray([np.dot(dagger(Vo_ni), Vo_ni)
                                  for Vo_ni in self.Vo_kni])

        if useibl:
            self.Fu_kii = self.s_lcao_kii - self.Fo_kii
        else:
            self.Vu_kni = [V_ni[M:self.N]
                           for V_ni, M in zip(self.V_kni, self.M_k)]
            self.Fu_kii = np.asarray([np.dot(dagger(Vu_ni), Vu_ni)
                                     for Vu_ni in self.Vu_kni])
        self.b_kil = []
        for Fu_ii, L in zip(self.Fu_kii, self.L_k):
            b_i, b_ii = la.eigh(Fu_ii)
            ls = b_i.real.argsort()[-L:]
            b_il = b_ii[:, ls]  # pick out the eigenvec with largest eigenvals.
            normalize2(b_il, Fu_ii)  # normalize the EDF: <phi_l|phi_l> = 1
            self.b_kil.append(b_il)

    def calculate_rotations(self):
        Uo_kni = [Vo_ni.copy() for Vo_ni in self.Vo_kni]
        Uu_kli = [np.dot(dagger(b_il), Fu_ii)
                  for b_il, Fu_ii in zip(self.b_kil, self.Fu_kii)]
        # Normalize such that <omega_i|omega_i> = <f_i|f_i>
        for Uo_ni, Uu_li, s_ii in zip(Uo_kni, Uu_kli, self.s_lcao_kii):
            normalize(Uo_ni, Uu_li, s_ii.diagonal())
        self.Uo_kni = Uo_kni
        self.Uu_kli = Uu_kli

    def calculate_overlaps(self):
        Wo_kii = [np.dot(dagger(Uo_ni), Uo_ni) for Uo_ni in self.Uo_kni]
        Wu_kii = [np.dot(dagger(Uu_li), Uu_li) for Uu_li in self.Uu_kli]
        # Wu_kii = [dots(dagger(Uu_li), dagger(b_il), Fu_ii, b_il, Uu_li)
        # for Uu_li, b_il, Fu_ii in zip(self.Uu_kli, self.b_kil, self.Fu_kii)]
        self.S_kii = np.asarray(Wo_kii) + np.asarray(Wu_kii)

    def get_condition_number(self):
        return np.asarray([condition_number(S) for S in self.S_kii])

    def calculate_hamiltonian_matrix(self, useibl=True):
        """Calculate H_kij = H^o_i(k)j(k) + H^u_i(k)j(k)
           i(k): Bloch sum of omega_i
        """

        epso_kn = [eps_n[:M] for eps_n, M in zip(self.eps_kn, self.M_k)]
        self.Ho_kii = np.asarray([np.dot(dagger(Uo_ni) * epso_n, Uo_ni)
                                  for Uo_ni, epso_n in zip(self.Uo_kni,
                                                           epso_kn)])

        if self.h_lcao_kii is not None and useibl:
            print("Using h_lcao and infinite band limit")
            Huf_kii = [h_lcao_ii - np.dot(dagger(Vo_ni) * epso_n, Vo_ni)
                       for h_lcao_ii, Vo_ni, epso_n in zip(self.h_lcao_kii,
                                                           self.Vo_kni,
                                                           epso_kn)]
            self.Huf_kii = np.asarray(Huf_kii)
        else:
            print("Using finite band limit (not using h_lcao)")
            epsu_kn = [eps_n[M:self.N]
                       for eps_n, M in zip(self.eps_kn, self.M_k)]
            Huf_kii = [np.dot(dagger(Vu_ni) * epsu_n, Vu_ni)
                       for Vu_ni, epsu_n in zip(self.Vu_kni, epsu_kn)]
            self.Huf_kii = np.asarray(Huf_kii)

        Hu_kii = [dots(dagger(Uu_li), dagger(b_il), Huf_ii, b_il, Uu_li)
                  for Uu_li, b_il, Huf_ii in zip(self.Uu_kli, self.b_kil,
                                                 self.Huf_kii)]
        self.Hu_kii = np.asarray(Hu_kii)
        self.H_kii = self.Ho_kii + self.Hu_kii

    def get_eigenvalues(self):
        return np.asarray([eigvals(H, S)
                           for H, S in zip(self.H_kii, self.S_kii)])

    def get_lcao_eigenvalues(self):
        return np.asarray([eigvals(H, S)
                           for H, S in zip(self.h_lcao_kii, self.s_lcao_kii)])

    def get_norm_of_projection(self):
        Sinv_kii = np.asarray([la.inv(S_ii) for S_ii in self.S_kii])

        normo_kn = np.asarray([dots(Uo_ni, Sinv_ii, dagger(Uo_ni)).diagonal()
                               for Uo_ni, Sinv_ii in
                               zip(self.Uo_kni, Sinv_kii)])

        Vu_kni = np.asarray([V_ni[M:self.N]
                             for V_ni, M in zip(self.V_kni, self.M_k)])

        Pu_kni = [dots(Vu_ni, b_il, Uu_li)
                  for Vu_ni, b_il, Uu_li in
                  zip(Vu_kni, self.b_kil, self.Uu_kli)]

        normu_kn = np.asarray([dots(Pu_ni, Sinv_ii, dagger(Pu_ni)).diagonal()
                               for Pu_ni, Sinv_ii in zip(Pu_kni, Sinv_kii)])

        return np.hstack((normo_kn, normu_kn))

    def calculate_functions(self, calc, basis, k=0):
        from gpaw.io import FileReference
        psit_nG = calc.wfs.kpt_u[k].psit_nG
        Uo_ni = self.Uo_kni[k]
        tarinstance = isinstance(psit_nG, FileReference)
        if tarinstance:
            psit_nG = np.asarray([psit_nG[i] for i in range(self.M_k[k])])

        basis_functions = get_bfs(calc)
        b_il, Uu_li, Vo_ni = self.b_kil[k], self.Uu_kli[k], self.Vo_kni[k]
        a_iG = -np.tensordot(Vo_ni, psit_nG, axes=[0, 0])  # a_iG = -fo_iG
        # self.fo_iG = -a_iG.copy()#f projected onto the occupied subspace
        C_iM = np.identity(self.Nw, dtype=self.dtype)
        basis_functions.lcao_to_grid(C_iM, a_iG, q=-1)  # a_iG=fu_iG=f_iG-fo_iG
        # self.f_iG = self.fo_iG + a_iG # check
        a_iG = np.tensordot(b_il, a_iG, axes=[0, 0])  # a_iG = EDF
        a_iG = np.tensordot(Uu_li, a_iG, axes=[0, 0])  # a_iG = wu_iG
        a_iG += np.tensordot(Uo_ni, psit_nG, axes=[0, 0])  # ai_G = wu_iG+wo_iG
        self.w_iG = a_iG

    def get_mlwf_initial_guess(self):
        """calculate initial guess for maximally localized
        wannier functions. Does not work for the infinite band limit.
        cu_nl: rotation coefficents of unoccupied states
        U_ii: rotation matrix of eigenstates and edf.
        """
        Vu_ni = self.Vu_ni[self.M: self.N]
        cu_nl = np.dot(Vu_ni, self.b_il)
        U_ii = np.vstack((self.Uo_ni, self.Uu_li))
        lowdin(U_ii)
        return U_ii, cu_nl



from gpaw.grid_descriptor import GridDescriptor
import pytest
import numpy as np


def test_symmetrize_fractional_translations():
    gd = GridDescriptor(N_c=np.array([5, 9, 6]))
    a_g = gd.empty(10)
    op_scc = [np.identity(3), -np.identity(3)]
    ft_sc = np.array([[0, 0, 0], [0.5, 0, 0]])
    with pytest.raises(ValueError, match=r"^The specified number"):
        gd.symmetrize(a_g, op_scc, ft_sc)

    newN_c = gd.get_nearest_compatible_grid(ft_sc)
    assert np.array_equal(newN_c, np.array([6, 9, 6]))
    ft2_sc = np.array([[1 / 20, 0, 0], [1 / 30, 1 / 4, 0]])
    newN_c = gd.get_nearest_compatible_grid(ft2_sc)
    assert np.array_equal(newN_c, np.array([60, 8, 6]))



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.
"""Main gpaw module."""
from __future__ import annotations
import os
import sys
import contextlib
from pathlib import Path
from typing import List, Union, Any, TYPE_CHECKING

__version__ = '24.6.0'
__ase_version_required__ = '3.23.0'

__all__ = ['GPAW',
           'Mixer', 'MixerSum', 'MixerDif', 'MixerSum2',
           'MixerFull',
           'CG', 'Davidson', 'RMMDIIS', 'DirectLCAO',
           'PoissonSolver',
           'FermiDirac', 'MethfesselPaxton', 'MarzariVanderbilt',
           'PW', 'LCAO', 'FD',
           'restart']

setup_paths: List[Union[str, Path]] = []
is_gpaw_python = '_gpaw' in sys.builtin_module_names
dry_run = 0

# When type-checking or running pytest, we want the debug-wrappers enabled:
debug: bool = (TYPE_CHECKING or
               'pytest' in sys.modules or
               bool(sys.flags.debug))


@contextlib.contextmanager
def disable_dry_run():
    """Context manager for temporarily disabling dry-run mode.

    Useful for skipping exit in the GPAW constructor.
    """
    global dry_run
    size = dry_run
    dry_run = 0
    yield
    dry_run = size


def get_scipy_version():
    import scipy
    # This is in a function because we don't like to have the scipy
    # import at module level
    return [int(x) for x in scipy.__version__.split('.')[:2]]


if 'OMP_NUM_THREADS' not in os.environ:
    os.environ['OMP_NUM_THREADS'] = '1'


class ConvergenceError(Exception):
    pass


class KohnShamConvergenceError(ConvergenceError):
    pass


class PoissonConvergenceError(ConvergenceError):
    pass


class KPointError(Exception):
    pass


class BadParallelization(Exception):
    """Error indicating missing parallelization support."""
    pass


def get_libraries() -> dict[str, str]:
    import gpaw.cgpaw as cgpaw
    libraries: dict[str, str] = {}
    if hasattr(cgpaw, 'lxcXCFunctional'):
        libraries['libxc'] = getattr(cgpaw, 'libxc_version', '2.x.y')
    else:
        libraries['libxc'] = ''
    return libraries


def parse_arguments(argv):
    from argparse import (ArgumentParser, REMAINDER,
                          RawDescriptionHelpFormatter)
    import warnings

    # With gpaw-python BLAS symbols are in global scope and we need to
    # ensure that NumPy and SciPy use symbols from their own dependencies
    if is_gpaw_python:
        old_dlopen_flags = sys.getdlopenflags()
        sys.setdlopenflags(old_dlopen_flags | os.RTLD_DEEPBIND)

    if is_gpaw_python:
        sys.setdlopenflags(old_dlopen_flags)

    version = sys.version.replace('\n', '')
    p = ArgumentParser(usage='%(prog)s [OPTION ...] [-c | -m] SCRIPT'
                       ' [ARG ...]',
                       description='Run a parallel GPAW calculation.\n\n'
                       f'Compiled with:\n  Python {version}',
                       formatter_class=RawDescriptionHelpFormatter)

    p.add_argument('--command', '-c', action='store_true',
                   help='execute Python string given as SCRIPT')
    p.add_argument('--module', '-m', action='store_true',
                   help='run library module given as SCRIPT')
    p.add_argument('-W', metavar='argument',
                   action='append', default=[], dest='warnings',
                   help='warning control.  See the documentation of -W for '
                   'the Python interpreter')
    p.add_argument('script', metavar='SCRIPT',
                   help='calculation script')
    p.add_argument('options', metavar='ARG',
                   help='arguments forwarded to SCRIPT', nargs=REMAINDER)

    args = p.parse_args(argv[1:])

    if args.command and args.module:
        p.error('-c and -m are mutually exclusive')

    sys.argv = [args.script] + args.options

    for w in args.warnings:
        # Need to convert between python -W syntax to call
        # warnings.filterwarnings():
        warn_args = w.split(':')
        assert len(warn_args) <= 5

        if warn_args[0] == 'all':
            warn_args[0] = 'always'
        if len(warn_args) >= 3:
            # e.g. 'UserWarning' (string) -> UserWarning (class)
            warn_args[2] = globals().get(warn_args[2])
        if len(warn_args) == 5:
            warn_args[4] = int(warn_args[4])

        warnings.filterwarnings(*warn_args, append=True)

    return args


def __getattr__(attr: str) -> Any:
    """
    Implement the lazy importing of classes in submodules."""

    import importlib

    try:
        import_target = all_lazy_imports[attr]
    except KeyError:
        raise AttributeError(
            f'{__getattr__.__module__}: no attribute named `.{attr!r}`'
        ) from None

    module, sep, target = import_target.rpartition('.')
    assert module and all(chunk.isidentifier() for chunk in module.split('.'))
    assert sep
    assert target.isidentifier()
    return getattr(importlib.import_module(module), target)


def __dir__() -> List[str]:
    """
    Get the normally-present module attributes and the lazily-imported objects.
    """
    return [*globals(), *all_lazy_imports]


all_lazy_imports = dict(
    OldGPAW='gpaw.calculator.GPAW',
    Mixer='gpaw.mixer.Mixer',
    MixerSum='gpaw.mixer.MixerSum',
    MixerDif='gpaw.mixer.MixerDif',
    MixerSum2='gpaw.mixer.MixerSum2',
    MixerFull='gpaw.mixer.MixerFull',

    Davidson='gpaw.eigensolvers.Davidson',
    RMMDIIS='gpaw.eigensolvers.RMMDIIS',
    CG='gpaw.eigensolvers.CG',
    DirectLCAO='gpaw.eigensolvers.DirectLCAO',

    PoissonSolver='gpaw.poisson.PoissonSolver',
    FermiDirac='gpaw.occupations.FermiDirac',
    MethfesselPaxton='gpaw.occupations.MethfesselPaxton',
    MarzariVanderbilt='gpaw.occupations.MarzariVanderbilt',
    FD='gpaw.wavefunctions.fd.FD',
    LCAO='gpaw.wavefunctions.lcao.LCAO',
    PW='gpaw.wavefunctions.pw.PW',
)


class BroadcastImports:
    def __enter__(self):
        from gpaw._broadcast_imports import broadcast_imports
        self._context = broadcast_imports
        return self._context.__enter__()

    def __exit__(self, *args):
        self._context.__exit__(*args)


broadcast_imports = BroadcastImports()


def main():
    with broadcast_imports:
        import runpy

        # Apparently we need the scipy.linalg import for compatibility?
        import scipy.linalg  # noqa: F401

        for attr in all_lazy_imports:
            __getattr__(attr)

        gpaw_args = parse_arguments(sys.argv)
    # The normal Python interpreter puts . in sys.path, so we also do that:
    sys.path.insert(0, '.')
    # Stacktraces can be shortened by running script with
    # PyExec_AnyFile and friends.  Might be nicer
    if gpaw_args.command:
        d = {'__name__': '__main__'}
        exec(gpaw_args.script, d, d)
    elif gpaw_args.module:
        # Python has: python [-m MOD] [-c CMD] [SCRIPT]
        # We use a much better way: gpaw-python [-m | -c] SCRIPT
        runpy.run_module(gpaw_args.script, run_name='__main__')
    else:
        runpy.run_path(gpaw_args.script, run_name='__main__')


if debug:
    import numpy as np
    np.seterr(over='raise', divide='raise', invalid='raise', under='ignore')
    oldempty = np.empty
    oldempty_like = np.empty_like

    def empty(*args, **kwargs):
        a = oldempty(*args, **kwargs)
        try:
            a.fill(np.nan)
        except ValueError:
            a.fill(42)
        return a

    def empty_like(*args, **kwargs):
        a = oldempty_like(*args, **kwargs)
        try:
            a.fill(np.nan)
        except ValueError:
            a.fill(-42)
        return a

    np.empty = empty
    np.empty_like = empty_like


def GPAW(*args, **kwargs) -> Any:
    if int(os.environ.get('GPAW_NEW', '0')):
        from gpaw.new.ase_interface import GPAW as NewGPAW
        return NewGPAW(*args, **kwargs)

    from gpaw.calculator import GPAW as OldGPAW
    return OldGPAW(*args, **kwargs)


def restart(filename, Class=None, **kwargs):
    if Class is None:
        from gpaw import GPAW as Class
    calc = Class(filename, **kwargs)
    atoms = calc.get_atoms()
    return atoms, calc


def read_rc_file():
    home = os.environ.get('HOME')
    if home is not None:
        rc = os.path.join(home, '.gpaw', 'rc.py')
        if os.path.isfile(rc):
            # Read file in ~/.gpaw/rc.py
            with open(rc) as fd:
                exec(fd.read())


def initialize_data_paths():
    try:
        setup_paths[:0] = os.environ['GPAW_SETUP_PATH'].split(os.pathsep)
    except KeyError:
        if len(setup_paths) == 0:
            if os.pathsep == ';':
                setup_paths[:] = [r'C:\gpaw-setups']
            else:
                setup_paths[:] = ['/usr/local/share/gpaw-setups',
                                  '/usr/share/gpaw-setups']


read_rc_file()
initialize_data_paths()


def RMM_DIIS(*args, **kwargs):
    import warnings
    from gpaw import RMMDIIS
    warnings.warn('Please use RMMDIIS instead of RMM_DIIS')
    return RMMDIIS(*args, **kwargs)



import os
from contextlib import contextmanager

import numpy as np
import pytest
from gpaw import setup_paths
from gpaw.cli.info import info
from gpaw.mpi import broadcast, world
from gpaw.test.gpwfile import GPWFiles, _all_gpw_methodnames
from gpaw.test.mmefile import MMEFiles
from gpaw.utilities import devnull


@contextmanager
def execute_in_tmp_path(request, tmp_path_factory):
    if world.rank == 0:
        # Obtain basename as
        # * request.function.__name__  for function fixture
        # * request.module.__name__    for module fixture
        basename = getattr(request, request.scope).__name__
        path = tmp_path_factory.mktemp(basename)
    else:
        path = None
    path = broadcast(path)
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(cwd)


@pytest.fixture(scope='function')
def in_tmp_dir(request, tmp_path_factory):
    """Run test function in a temporary directory."""
    with execute_in_tmp_path(request, tmp_path_factory) as path:
        yield path


@pytest.fixture(scope='module')
def module_tmp_path(request, tmp_path_factory):
    """Run test module in a temporary directory."""
    with execute_in_tmp_path(request, tmp_path_factory) as path:
        yield path


@pytest.fixture
def add_cwd_to_setup_paths():
    """Temporarily add current working directory to setup_paths."""
    try:
        setup_paths[:0] = ['.']
        yield
    finally:
        del setup_paths[:1]


@pytest.fixture(scope='session')
def sessionscoped_monkeypatch():
    # The standard monkeypatch fixture is function scoped
    # so we need to roll our own
    with pytest.MonkeyPatch.context() as monkeypatch:
        yield monkeypatch


@pytest.fixture(autouse=True, scope='session')
def monkeypatch_response_spline_points(sessionscoped_monkeypatch):
    import gpaw.response.paw as paw
    # https://gitlab.com/gpaw/gpaw/-/issues/984
    sessionscoped_monkeypatch.setattr(paw, 'DEFAULT_RADIAL_POINTS', 2**10)


@pytest.fixture(scope='session')
def gpw_files(request):
    """Reuse gpw-files.

    Returns a dict mapping names to paths to gpw-files.
    The files are written to the pytest cache and can be cleared using
    pytest --cache-clear.

    Example::

        def test_something(gpw_files):
            calc = GPAW(gpw_files['h2_lcao'])
            ...

    Possible systems are:

    * Bulk BCC-Li with 3x3x3 k-points: ``bcc_li_pw``, ``bcc_li_fd``,
      ``bcc_li_lcao``.

    * O2 molecule: ``o2_pw``.

    * H2 molecule: ``h2_pw``, ``h2_fd``, ``h2_lcao``.

    * H2 molecule (not centered): ``h2_pw_0``.

    * N2 molecule ``n2_pw``

    * N molecule ``n_pw``

    * Spin-polarized H atom: ``h_pw``.

    * Polyethylene chain.  One unit, 3 k-points, no symmetry:
      ``c2h4_pw_nosym``.  Three units: ``c6h12_pw``.

    * Bulk BN (zinkblende) with 2x2x2 k-points and 9 converged bands:
      ``bn_pw``.

    * h-BN layer with 3x3x1 (gamma center) k-points and 26 converged bands:
      ``hbn_pw``.

    * Graphene with 6x6x1 k-points: ``graphene_pw``

    * I2Sb2 (Z2 topological insulator) with 6x6x1 k-points and no
      symmetries: ``i2sb2_pw_nosym``

    * MoS2 with 6x6x1 k-points: ``mos2_pw`` and ``mos2_pw_nosym``

    * MoS2 with 5x5x1 k-points: ``mos2_5x5_pw``

    * NiCl2 with 6x6x1 k-points: ``nicl2_pw`` and ``nicl2_pw_evac``

    * V2Br4 (AFM monolayer), LDA, 4x2x1 k-points, 28(+1) converged bands:
      ``v2br4_pw`` and ``v2br4_pw_nosym``

    * Bulk Si, LDA, 2x2x2 k-points (gamma centered): ``si_pw``

    * Bulk Si, LDA, 4x4x4 k-points, 8(+1) converged bands: ``fancy_si_pw``
      and ``fancy_si_pw_nosym``

    * Bulk SiC, LDA, 4x4x4 k-points, 8(+1) converged bands: ``sic_pw``
      and ``sic_pw_spinpol``

    * Bulk Fe, LDA, 4x4x4 k-points, 9(+1) converged bands: ``fe_pw``
      and ``fe_pw_nosym``

    * Bulk C, LDA, 2x2x2 k-points (gamma centered), ``c_pw``

    * Bulk Co (HCP), 4x4x4 k-points, 12(+1) converged bands: ``co_pw``
      and ``co_pw_nosym``

    * Bulk SrVO3 (SC), 3x3x3 k-points, 20(+1) converged bands: ``srvo3_pw``
      and ``srvo3_pw_nosym``

    * Bulk Al, LDA, 4x4x4 k-points, 10(+1) converged bands: ``al_pw``
      and ``al_pw_nosym``

    * Bulk Al, LDA, 4x4x4 k-points, 4 converged bands: ``bse_al``

    * Bulk Ag, LDA, 2x2x2 k-points, 6 converged bands,
      2eV U on d-band: ``ag_pw``

    * Bulk GaAs, LDA, 4x4x4 k-points, 8(+1) bands converged: ``gaas_pw``
      and ``gaas_pw_nosym``

    * Bulk P4, LDA, 4x4 k-points, 40 bands converged: ``p4_pw``

    * Distorted bulk Fe, revTPSS: ``fe_pw_distorted``

    * Distorted bulk Si, TPSS: ``si_pw_distorted``

    Files always include wave functions.
    """
    cache = request.config.cache
    gpaw_cachedir = cache.mkdir('gpaw_test_gpwfiles')

    gpwfiles = GPWFiles(gpaw_cachedir)

    try:
        setup_paths.append(gpwfiles.testing_setup_path)
        yield gpwfiles
    finally:
        setup_paths.remove(gpwfiles.testing_setup_path)


@pytest.fixture(scope='session', params=sorted(_all_gpw_methodnames))
def all_gpw_files(request, gpw_files, pytestconfig):
    """This fixture parametrizes a test over all gpw_files.

    For example pytest test_generate_gpwfiles.py -n 16 is a way to quickly
    generate all gpw files independently of the rest of the test suite."""

    # Note: Parametrizing over _all_gpw_methodnames must happen *after*
    # it is populated, i.e., further down in the file than
    # the @gpwfile decorator.

    import os
    gpaw_new = int(os.environ.get('GPAW_NEW', '0'))

    # TODO This xfail-information should probably live closer to the
    # gpwfile definitions and not here in the fixture.
    skip_if_new = {'Cu3Au_qna',
                   'nicl2_pw', 'nicl2_pw_evac',
                   'v2br4_pw', 'v2br4_pw_nosym',
                   'sih4_xc_gllbsc_fd', 'sih4_xc_gllbsc_lcao',
                   'na2_isolated', 'h2o_xas'}
    if gpaw_new and request.param in skip_if_new:
        pytest.xfail(f'{request.param} gpwfile not yet working with GPAW_NEW')

    # Accessing each file via __getitem__ executes the calculation:
    return gpw_files[request.param]


@pytest.fixture(scope='session')
def mme_files(request, gpw_files):
    """Reuse mme files"""
    cache = request.config.cache
    mme_cachedir = cache.mkdir('gpaw_test_mmefiles')

    return MMEFiles(mme_cachedir, gpw_files)


class GPAWPlugin:
    def __init__(self):
        if world.rank == -1:
            print()
            info()

    def pytest_terminal_summary(self, terminalreporter, exitstatus, config):
        from gpaw.mpi import size
        terminalreporter.section('GPAW-MPI stuff')
        terminalreporter.write(f'size: {size}\n')


@pytest.fixture
def sg15_hydrogen():
    from io import StringIO
    from gpaw.test.pseudopotential.H_sg15 import pp_text
    from gpaw.upf import read_sg15
    # We can't easily load a non-python file from the test suite.
    # Therefore we load the pseudopotential from a Python file.
    return read_sg15(StringIO(pp_text))


def pytest_configure(config):
    # Allow for fake cupy:
    os.environ['GPAW_CPUPY'] = '1'

    if world.rank != 0:
        try:
            tw = config.get_terminal_writer()
        except (AssertionError, AttributeError):
            pass
        else:
            tw._file = devnull
    config.pluginmanager.register(GPAWPlugin(), 'pytest_gpaw')


def pytest_runtest_setup(item):
    """Skip some tests.

    If:

    * they depend on libxc and GPAW is not compiled with libxc
    * they are before $PYTEST_START_AFTER
    """
    from gpaw import get_libraries
    libraries = get_libraries()

    if world.size > 1:
        for mark in item.iter_markers():
            if mark.name == 'serial':
                pytest.skip('Only run in serial')

    if item.location[0] <= os.environ.get('PYTEST_START_AFTER', ''):
        pytest.skip('Not after $PYTEST_START_AFTER')
        return

    if libraries['libxc']:
        return

    if any(mark.name in {'libxc', 'mgga'}
           for mark in item.iter_markers()):
        pytest.skip('No LibXC.')


@pytest.fixture
def scalapack():
    """Skip if not compiled with sl.

    This fixture otherwise does not return or do anything."""
    from gpaw.utilities import compiled_with_sl
    if not compiled_with_sl():
        pytest.skip('no scalapack')


@pytest.fixture
def needs_ase_master():
    from ase.utils.filecache import MultiFileJSONCache
    try:
        MultiFileJSONCache('bla-bla', comm=None)
    except TypeError:
        pytest.skip('ASE is too old')


def pytest_report_header(config, startdir):
    # Use this to add custom information to the pytest printout.
    yield f'GPAW MPI rank={world.rank}, size={world.size}'

    # We want the user to be able to see where gpw files are cached,
    # but the only way to see the cache location is to make a directory
    # inside it.  mkdir('') returns the toplevel cache dir without
    # actually creating a subdirectory:
    cachedir = config.cache.mkdir('')
    yield f'Cache directory including gpw files: {cachedir}'


@pytest.fixture
def rng():
    """Seeded random number generator.

    Tests should be deterministic and should use this
    fixture or initialize their own rng."""
    return np.random.default_rng(42)


@pytest.fixture
def gpaw_new() -> bool:
    """Are we testing the new code?"""
    return bool(int(os.environ.get('GPAW_NEW', '0')))



from gpaw.directmin.locfunc.etdm_localization_fdpw import FDPWETDMLocalize
from gpaw.directmin.fdpw.er_localization import ERLocalization as ERL
from gpaw.directmin.functional.fdpw import get_functional \
    as get_functional_fdpw
from gpaw.pipekmezey.pipek_mezey_wannier import PipekMezey
from gpaw.pipekmezey.wannier_basic import WannierLocalization
import numpy as np


def localize_orbitals(
        wfs, dens, ham, log, localizationtype, tol=None, seed=None,
        func_settings=None):
    io = localizationtype

    if io is None:
        return

    locnames = io.lower().split('_')
    if tol is not None:
        tols = {name: tol for name in locnames}
    else:
        tols = {'er': 5.0e-5,
                'pm': 1.0e-10,
                'fb': 1.0e-10,
                'pz': 5.0e-4,
                'ks': 5.0e-4}

    log("Initial localization: ...", flush=True)
    wfs.timer.start('Initial Localization')
    for name in locnames:
        tol = tols[name]
        if name == 'er':
            if wfs.mode == 'lcao':
                log('Edmiston-Ruedenberg localization is not supported '
                    'in LCAO',
                    flush=True)
                continue
            log('Edmiston-Ruedenberg localization started',
                flush=True)
            dm = FDPWETDMLocalize(
                ERL(wfs, dens), wfs,
                maxiter=200, g_tol=tol, randval=0.1)
            dm.run(wfs, dens, seed=seed)
            log('Edmiston-Ruedenberg localization finished',
                flush=True)
            del dm
        elif name == 'pz':
            if wfs.mode != 'lcao':
                log('Perdew-Zunger localization started', flush=True)
                PZC = get_functional_fdpw(func_settings, wfs, dens, ham)
                dm = FDPWETDMLocalize(
                    PZC, wfs, maxiter=200, g_tol=tol, randval=0.1)
                dm.run(wfs, dens, log=log)
                log('Perdew-Zunger localization finished', flush=True)
        elif name == 'ks':
            log('ETDM minimization using occupied and virtual orbitals',
                flush=True)
            if wfs.mode == 'lcao':
                raise NotImplementedError
            else:
                KS = get_functional_fdpw('ks', wfs, dens, ham)
            dm = FDPWETDMLocalize(KS, wfs, maxiter=200, g_tol=tol, randval=0)
            dm.run(wfs, dens, ham=ham, log=log)
            log('ETDM minimization finished', flush=True)
        else:
            for k, kpt in enumerate(wfs.kpt_u):
                if sum(kpt.f_n) < 1.0e-3:
                    continue
                if name == 'pm':
                    if k == 0:
                        log('Pipek-Mezey localization started',
                            flush=True)
                    lf_obj = PipekMezey(
                        wfs=wfs, spin=kpt.s, dtype=wfs.dtype, seed=seed)
                    lf_obj.localize(tolerance=tol)
                    if k == 0:
                        log('Pipek-Mezey localization finished',
                            flush=True)
                    U = np.ascontiguousarray(
                        lf_obj.W_k[kpt.q].T)
                elif name == 'fb':
                    if k == 0:
                        log('Foster-Boys localization started',
                            flush=True)
                    lf_obj = WannierLocalization(
                        wfs=wfs, spin=kpt.s, seed=seed)
                    lf_obj.localize(tolerance=tol)
                    if k == 0:
                        log('Foster-Boys localization finished',
                            flush=True)
                    U = np.ascontiguousarray(
                        lf_obj.U_kww[kpt.q].T)
                    if wfs.dtype == float:
                        U = U.real
                else:
                    raise ValueError('Check localization type.')
                wfs.gd.comm.broadcast(U, 0)
                dim = U.shape[0]
                if wfs.mode == 'fd':
                    kpt.psit_nG[:dim] = np.einsum(
                        'ij,jkml->ikml', U, kpt.psit_nG[:dim])
                elif wfs.mode == 'pw':
                    kpt.psit_nG[:dim] = U @ kpt.psit_nG[:dim]
                else:
                    kpt.C_nM[:dim] = U @ kpt.C_nM[:dim]

                del lf_obj

    wfs.timer.stop('Initial Localization')
    log("Done", flush=True)



from ase.build import molecule
from ase.parallel import parprint

from gpaw import GPAW
from gpaw.utilities.adjust_cell import adjust_cell
from gpaw.analyse.hirshfeld import HirshfeldDensity, HirshfeldPartitioning
from gpaw.analyse.wignerseitz import WignerSeitz
import pytest


@pytest.mark.later
def test_utilities_partitioning(in_tmp_dir):
    h = 0.4
    gpwname = 'H2O' + str(h) + '.gpw'

    def run(lastres=[]):
        results = []

        # Hirshfeld ----------------------------------------

        if 1:

            hd = HirshfeldDensity(calc)

            # check for the number of electrons
            expected = [[None, 10],
                        [[0, 1, 2], 10],
                        [[1, 2], 2],
                        [[0], 8],
                        ]
            for gridrefinement in [1, 2, 4]:
                # Test for all gridrefinements for get_all_electron_density
                parprint('grid refinement', gridrefinement)
                for result in expected:
                    indicees, result = result
                    full, gd = hd.get_density(indicees, gridrefinement)
                    parprint('indicees', indicees, end=': ')
                    parprint('result, expected:', gd.integrate(full), result)
                    if gridrefinement < 4:
                        # The highest level of gridrefinement gets wrong
                        # electron numbers
                        assert gd.integrate(full) == pytest.approx(result,
                                                                   abs=1.e-8)
                    else:
                        assert gd.integrate(full) == pytest.approx(result,
                                                                   abs=1.e-4)

            hp = HirshfeldPartitioning(calc)
            vr = hp.get_effective_volume_ratios()
            parprint('Hirshfeld:', vr)
            if len(lastres):
                assert vr == pytest.approx(lastres.pop(0), abs=1.e-10)
            results.append(vr)

        # Wigner-Seitz ----------------------------------------

        if 1:
            ws = WignerSeitz(calc.density.finegd, mol, calc)

            vr = ws.get_effective_volume_ratios()
            parprint('Wigner-Seitz:', vr)
            if len(lastres):
                assert vr == pytest.approx(lastres.pop(0), abs=1.e-10)
            results.append(vr)

        return results

    mol = molecule('H2O')
    adjust_cell(mol, 2.5, h=h)

    # calculate
    if 1:
        parprint('### fresh:')
        calc = GPAW(mode='fd',
                    nbands=6,
                    h=h,
                    txt=None)
    if 1:
        calc.calculate(mol)
        calc.write(gpwname)
        lastres = run()

    # load previous calculation
    if 1:
        parprint('### reloaded:')
        calc = GPAW(gpwname, txt=None)
        mol = calc.get_atoms()
        run(lastres)

    # periodic modulo test
    parprint('### periodic:')
    mol.set_pbc(True)
    mol.translate(-mol[0].position)
    mol.translate([-1.e-24, 0, 0])
    calc.calculate(mol)
    run()



from functools import cached_property
import numpy as np
import pytest

from gpaw.test import findpeak
from gpaw.response.df import DielectricFunction
from ase.units import Bohr, Hartree


class Helper:
    def __init__(self, gpw, integrationmode):
        self._gpw = gpw
        self._integrationmode = integrationmode

    @cached_property
    def df(self):
        return DielectricFunction(
            self._gpw,
            frequencies={'type': 'nonlinear',
                         'domega0': 0.03},
            ecut=10,
            rate=0.1,
            integrationmode=self._integrationmode,
            txt=None)

    @cached_property
    def lfc(self):
        # lfc == local field corrections (?)
        return {axis: self.df.get_dielectric_function(direction=axis)[1]
                for axis in 'xyz'}

    @cached_property
    def wp(self):
        chi0_drude = self.df.chi0calc.chi0_opt_ext_calc.drude_calc.calculate(
            self.df.chi0calc.chi0_opt_ext_calc.wd, 0.1)
        return chi0_drude.plasmafreq_vv[0, 0]**0.5

    @cached_property
    def w_w(self):
        return self.df.chi0calc.chi0_opt_ext_calc.wd.omega_w

    def _compare_peak(self, calc, axis):
        df1LFCx = self.lfc[axis]
        df2LFCx = calc.lfc[axis]
        # w_x equal for paired & polarized tetra
        w1, I1 = findpeak(self.w_w, -(1. / df1LFCx).imag)
        w2, I2 = findpeak(self.w_w, -(1. / df2LFCx).imag)
        assert w1 == pytest.approx(w2, abs=1e-3)
        assert I1 == pytest.approx(I2, abs=0.1)

    def compare_peaks(self, calc):
        for axis in 'xyz':
            self._compare_peak(calc, axis)


@pytest.mark.dielectricfunction
@pytest.mark.tetrahedron
@pytest.mark.response
def test_chi0_intraband(in_tmp_dir, gpw_files):
    """Comparing the plasmon peaks found in bulk sodium for two different
    atomic structures. Testing for idential plasmon peaks. Not using
    physical sodium cell."""
    intraband_spinpaired = gpw_files['intraband_spinpaired_fulldiag']
    intraband_spinpolarized = gpw_files['intraband_spinpolarized_fulldiag']

    calc1 = Helper(intraband_spinpaired, 'tetrahedron integration')
    calc2 = Helper(intraband_spinpaired, None)
    calc3 = Helper(intraband_spinpolarized, 'tetrahedron integration')
    calc4 = Helper(intraband_spinpolarized, None)

    # Compare plasmon frequencies and intensities
    w_w = calc1.w_w

    # frequency grids must be the same
    for calc in [calc1, calc2, calc3, calc4]:
        assert np.allclose(calc.w_w, w_w, atol=1e-5, rtol=1e-4)

    # Analytical Drude result
    n = 1 / (calc1.df.gs.volume * Bohr**-3)
    drude_wp = np.sqrt(4 * np.pi * n)

    # From https://doi.org/10.1021/jp810808h
    ref_wp = 5.71 / Hartree

    # spin paired matches spin polar - tetra
    assert calc1.wp == pytest.approx(calc3.wp, abs=1e-2)
    # spin paired matches spin polar - none
    assert calc2.wp == pytest.approx(calc4.wp, abs=1e-2)
    # Use larger margin when comparing to Drude
    assert calc1.wp == pytest.approx(drude_wp, abs=0.5)
    # Use larger margin when comparing to Drude
    assert calc2.wp == pytest.approx(drude_wp, abs=0.5)
    # paired tetra match paper
    assert calc1.wp == pytest.approx(ref_wp, abs=0.1)
    # paired none match paper
    assert calc2.wp == pytest.approx(ref_wp, abs=0.1)

    # w_x, w_y and w_z equal for paired & polarized tetra
    calc1.compare_peaks(calc3)

    # w_x, w_y and w_z equal for paired & polarized none
    calc2.compare_peaks(calc4)



r"""This module calculates the orbital magnetic moment vector for each atom.

The orbital magnetic moment is calculated in the atom-centred approximation
where only the PAW correction to the wave function is assumed to contribute.
This leads to the equation (presented in SI units):

::

                 ===
   a         e   \     a    a
  m      = - --  /    D    L
   orb,v     2m  ===   ii'  vii'
                 ii'

with L^a_vii' containing the matrix elements of the angular momentum operator
between two partial waves centred at atom a.

The orbital magnetic moments are returned in units of μ_B without the sign of
the negative electronic charge, q = - e.
"""

import numpy as np
from gpaw.new import zips
from gpaw.spinorbit import get_L_vlmm

L_vlmm = get_L_vlmm()


def calculate_orbmag_from_density(D_asii, n_aj, l_aj):
    """Returns orbital magnetic moment vectors for each atom a
    calculated from its respective atomic density matrix.

    This method assumes that D_asii is on every rank and not parallelised.

    Parameters
    ----------
    D_asii : AtomArrays or dictionary
        Atomic density matrix for each atom a. The i-index refers to the
        partial waves of an atom and the s-index refers to 0, x, y, and z.
    n_aj : List of lists of integers
        Principal quantum number for each radial partial wave j
    l_aj : List of lists of integers
        Angular momentum quantum number for each radial partial wave j

    NB: i is an index for all partial waves for one atom and j is an index for
    only the radial wave function which is used to build all of the partial
    waves. i and j do not refer to the same kind of index.

    Only pairs of partial waves with the same radial function may yield
    nonzero contributions. The sum can therefore be limited to diagonal blocks
    of shape [2 * l_j + 1, 2 * l_j + 1] where l_j is the angular momentum
    quantum number of the j'th radial function.

    Partials with unbounded radial functions (negative n_j) are skipped.
    """

    orbmag_av = np.zeros([len(n_aj), 3])
    for (a, D_sii), n_j, l_j in zips(D_asii.items(), n_aj, l_aj):
        assert D_sii.shape[0] == 4
        D_ii = D_sii[0]  # Only the electron density

        Ni = 0
        for n, l in zips(n_j, l_j):
            Nm = 2 * l + 1
            if n < 0:
                Ni += Nm
                continue
            for v in range(3):
                orbmag_av[a, v] += np.einsum('ij,ij->',
                                             D_ii[Ni:Ni + Nm, Ni:Ni + Nm],
                                             L_vlmm[v][l]).real
            Ni += Nm
    return orbmag_av



from math import sqrt
import pytest

from ase import Atoms

from gpaw.cluster import Cluster


def test_CO(recwarn):
    R = 2.0
    box = 4.
    h = 0.2
    CO = Cluster(Atoms('CO', [(1, 0, 0), (1, 0, R)]))

    CO.rotate(90, 'y')
    assert CO.positions[1, 0] == pytest.approx(R, abs=1e-10)

    # translate
    CO.translate(-CO.get_center_of_mass())
    p = CO.positions.copy()
    for i in range(2):
        assert p[i, 1] == pytest.approx(0, abs=1e-10)
        assert p[i, 2] == pytest.approx(0, abs=1e-10)

    CO.rotate(p[1] - p[0], (1, 1, 1))
    q = CO.positions.copy()
    for c in range(3):
        assert q[0, c] == pytest.approx(p[0, 0] / sqrt(3), abs=1e-10)
        assert q[1, c] == pytest.approx(p[1, 0] / sqrt(3), abs=1e-10)

    CO.minimal_box(box, h)
    w = recwarn.pop(FutureWarning)

    assert (str(w.message) ==
            'Please use adjust_cell from gpaw.utilities.adjust_cell instead.')

    CO.find_connected(0)
    w = recwarn.pop(FutureWarning)

    assert (str(w.message) ==
            'Please use connected_indices from ase.build.connected instead.')



from typing import List, Tuple

import numpy as np
from ase.units import Ha

from gpaw.typing import Array1D, Array2D, Array3D, ArrayLike2D


def parse_hubbard_string(type: str) -> Tuple[str, 'HubbardU']:

    # Parse DFT+U parameters from type-string:
    # Examples: "type:l,U" or "type:l,U,scale"
    type, lus = type.split(':')
    if type == '':
        type = 'paw'

    l = []
    U = []
    scale = []

    for lu in lus.split(';'):  # Multiple U corrections
        l_, u_, scale_ = (lu + ',,').split(',')[:3]
        l.append('spdf'.find(l_))
        U.append(float(u_) / Ha)
        if scale_:
            scale.append(bool(int(scale_)))
        else:
            scale.append(True)
    return type, HubbardU(U, l, scale)


class HubbardU:
    def __init__(self, U, l, scale=1):
        self.scale = scale
        self.U = U
        self.l = l

    def _tuple(self):
        # Tests use this method to compare to expected values
        return (self.l, self.U, self.scale)

    def calculate(self, setup, D_sii):
        eU = 0.
        dHU_sii = np.zeros_like(D_sii)
        for l, U, scale in zip(self.l, self.U, self.scale):
            nl = np.argwhere(np.equal(setup.l_j, l))
            if not (len(nl) == 1 or len(nl) == 2):
                raise RuntimeError(f'Setup has {len(nl)} radial partial waves '
                                   f'with angular momentum quantum number {l}.'
                                   ' Must be 1 or 2 for DFT+U.')
            if (len(nl) == len(np.argwhere(np.equal(np.array(setup.n_j)[nl],
                                                    -1))) and scale == 1):
                raise RuntimeError('DFT+U correction cannot be scaled if '
                                   'there is no bounded partial waves.')

            eU1, dHU1_sii = hubbard(D_sii, U=U, l=l,
                                    l_j=setup.l_j, n_j=setup.n_j,
                                    N0_q=setup.N0_q, scale=scale)
            eU += eU1.real
            dHU_sii += dHU1_sii
        return eU, dHU_sii

    def descriptions(self):
        for U, l, scale in zip(self.U, self.l, self.scale):
            yield f'Hubbard: {{U: {U * Ha},  # eV\n'
            yield f'          l: {l},\n'
            yield f'          scale: {bool(scale)}}}'


def hubbard(D_sii: Array3D,
            U: float,
            l: int,
            l_j: List[int],
            n_j: List[int],
            N0_q: Array1D,
            scale: bool) -> Tuple[float, ArrayLike2D]:
    nspins = len(D_sii)

    # j-indices which have the correct angular momentum quantum number
    nl = np.where(np.equal(l_j, l))[0]

    nm_j = 2 * np.array(l_j) + 1
    nm = nm_j[nl[0]]

    # Get relevant entries of the density matrix
    i1 = slice(nm_j[:nl[0]].sum(), nm_j[:nl[0]].sum() + nm)

    eU = 0.
    dHU_sii = np.zeros_like(D_sii)

    for s, D_ii in enumerate(D_sii):
        N_mm, dHU_ii = aoom(D_ii, l, l_j, n_j, N0_q, scale)
        N_mm = N_mm / 2 * nspins

        if nspins == 4:
            N_mm = N_mm / 2.0
            if s == 0:
                eU1 = U / 2. * (N_mm - 0.5 * N_mm @ N_mm).trace()

                dHU_mm = U / 2. * (np.eye(nm) - N_mm.T)

            else:
                eU1 = -U / 2. * (0.5 * N_mm @ N_mm).trace()

                dHU_mm = -U / 2. * N_mm.T
        else:
            eU1 = U / 2. * (N_mm - N_mm @ N_mm).trace()

            dHU_mm = U / 2. * (np.eye(nm) - 2 * N_mm)

        eU += eU1
        if nspins == 1:
            # Add contribution from other spin manifold
            eU += eU1

        if len(nl) == 1:
            dHU_ii[i1, i1] *= dHU_mm
        elif len(nl) == 2:
            i2 = slice(nm_j[:nl[1]].sum(), nm_j[:nl[1]].sum() + nm)

            dHU_ii[i1, i1] *= dHU_mm
            dHU_ii[i1, i2] *= dHU_mm
            dHU_ii[i2, i1] *= dHU_mm
            dHU_ii[i2, i2] *= dHU_mm
        else:
            raise NotImplementedError

        dHU_sii[s] = dHU_ii

    return eU, dHU_sii


def aoom(D_ii: Array2D,
         l: int,
         l_j: List[int],
         n_j: List[int],
         N0_q: Array1D,
         scale: bool = True) -> Tuple[Array2D, Array2D]:
    """
    This function returns the atomic orbital occupation matrix (aoom) for a
    given l quantum number.

    The submatrix / submatrices of the density matrix (D_ii) for the
    selected l quantum number are determined and summed together which
    represents the orbital occupation matrix. For l=2 this is a 5x5 matrix.

    If scale = True, the inner products are scaled such that the inner product
    of the bounded partial waves is 1.
    """

    # j-indices which have the correct angular momentum quantum number
    nl = np.where(np.equal(l_j, l))[0]

    nm_j = 2 * np.array(l_j) + 1
    nm = nm_j[nl[0]]

    # Get relevant entries of the density matrix
    i1 = slice(nm_j[:nl[0]].sum(), nm_j[:nl[0]].sum() + nm)

    dHU_ii = np.zeros_like(D_ii)
    if len(nl) == 2:
        # First get q-indices for the radial inner products
        q1 = nl[0] * len(l_j) - (nl[0] - 1) * nl[0] // 2  # Bounded-bounded
        q2 = nl[1] * len(l_j) - (nl[1] - 1) * nl[1] // 2  # Unbounded-unbounded
        q12 = q1 + nl[1] - nl[0]  # Bounded-unbounded

        # If the Hubbard correction should be scaled, the three inner products
        # will be divided by the inner product of the bounded partial wave,
        # increasing the value of these inner products since 0 < N0_q[q1] < 1
        if scale:
            if n_j[nl[1]] == -1:
                N1 = 1
                N2 = N0_q[q2] / N0_q[q1]
                N12 = N0_q[q12] / N0_q[q1]
            else:
                N1 = 1
                N2 = 1
                N12 = N0_q[q12] / np.sqrt(N0_q[q1] * N0_q[q2])
        else:
            N1 = N0_q[q1]
            N2 = N0_q[q2]
            N12 = N0_q[q12]

        # Get the entries in the density matrix of the unbounded partial waves
        i2 = slice(nm_j[:nl[1]].sum(), nm_j[:nl[1]].sum() + nm)

        # Scale and add the four submatrices to get the occupation matrix
        N_mm = D_ii[i1, i1] * N1 + D_ii[i2, i2] * N2 + (D_ii[i1, i2]
                                                        + D_ii[i2, i1]) * N12

        dHU_ii[i1, i1] = N1
        dHU_ii[i1, i2] = N12
        dHU_ii[i2, i1] = N12
        dHU_ii[i2, i2] = N2

        return N_mm, dHU_ii
    elif len(nl) == 1:
        q1 = nl[0] * len(l_j) - (nl[0] - 1) * nl[0] // 2
        if scale:
            N1 = 1
        else:
            N1 = N0_q[q1]

        N_mm = D_ii[i1, i1] * N1
        dHU_ii[i1, i1] = N1
        return N_mm, dHU_ii
    else:
        raise NotImplementedError(f'Setup has {len(nl)} partial waves with '
                                  f'angular momentum quantum number {l}. '
                                  'Must be 1 or 2 for DFT+U correction.')



"""Unit test for elph/supercell and elph/gmatrix checking
whether they properly respect the self.indices.

Note: If the users gives an index twice that's passed through, because that's
      how ase.phonons does it.

"""
import numpy as np
import pytest

from ase.build import bulk
# from ase.phonons import Phonons
from gpaw.elph import Supercell, ElectronPhononMatrix
# from gpaw.mpi import world


class FakeElectronPhononMatrix(ElectronPhononMatrix):
    def _set_supercell_cache(self, supercell_cache: str,
                             load_sc_as_needed: bool):
        pass

    def _read_phonon_cache(self):
        pass


def _supercell_all(atoms):
    natoms = len(atoms)
    sc = Supercell(atoms)
    assert len(sc.indices) == natoms
    assert sc.indices[0] == 0
    assert sc.indices[-1] == natoms - 1


def _gmatrix_all(atoms):
    natoms = len(atoms)
    fepm = FakeElectronPhononMatrix(atoms, None, {"name": "phonon"})
    assert len(fepm.indices) == natoms
    assert fepm.indices[0] == 0
    assert fepm.indices[-1] == natoms - 1


def _supercell_indices(atoms, indices):
    natoms = len(atoms)
    nindices = len(indices)
    sci = Supercell(atoms, indices=indices)
    assert len(sci.indices) != natoms
    assert len(sci.indices) == nindices
    assert sci.indices[0] == indices[0]
    assert sci.indices[-1] == indices[-1]


def _gmatrix_indices(atoms, indices):
    natoms = len(atoms)
    nindices = len(indices)
    fepm = FakeElectronPhononMatrix(atoms, None, {"name": "phonon"},
                                    indices=indices)
    assert len(fepm.indices) != natoms
    assert len(fepm.indices) == nindices
    assert fepm.indices[0] == indices[0]
    assert fepm.indices[-1] == indices[-1]


@pytest.mark.elph
def test_indices():

    # arbitrary atoms object
    n = 3
    atoms = bulk('Li', crystalstructure='bcc', a=3.51, cubic=True) * (n, n, n)
    assert len(atoms) == 2 * n**3

    # case 1: self.indices not given
    _supercell_all(atoms)
    _gmatrix_all(atoms)

    # TODO: Supercell.calculate_gradient()? Would need fake cache.
    # TODO: Supercell.calculate_supercell_matrix()? Would need fake cache and
    #       stuff.

    # case 2: gives some random indices
    seed = 123456
    rng = np.random.default_rng(seed)
    indices = rng.integers(0, len(atoms),
                           rng.integers(0, len(atoms) - 1))
    _supercell_indices(atoms, indices)
    _gmatrix_indices(atoms, indices)
    # TODO: Supercell.calculate_gradient()? Would need fake cache.
    # TODO: Supercell.calculate_supercell_matrix()? Would need fake cache
    #       and stuff.



import pytest
from gpaw.response.mpa_sampling import mpa_frequency_sampling


@pytest.mark.response
def test_mpa_sampling_errors():
    with pytest.raises(ValueError):
        mpa_frequency_sampling(2, wrange=[0, 0], varpi=1,
                               eta0=0.1, eta_rest=0.1, parallel_lines=3,
                               alpha=1)
    with pytest.raises(ValueError):
        mpa_frequency_sampling(2, wrange=[0, 0], varpi=1,
                               eta0=0.1, eta_rest=0.1, parallel_lines=1,
                               alpha=1)


@pytest.mark.response
def test_mpa_sampling_1pole():
    # print("npol=1, parallel_lines=1, w1=0.1j, w2=1j, alpha=1:")
    with pytest.raises(AssertionError):
        mpa_frequency_sampling(1, wrange=[0, 0], varpi=1,
                               eta0=0.1, eta_rest=0.1, parallel_lines=1,
                               alpha=1)

    # print("npol=1, parallel_lines=2, w1=0.1j, w2=1j, alpha=0:")
    w_grid = mpa_frequency_sampling(1, wrange=[0, 0], varpi=1,
                                    eta0=0.1, eta_rest=0.1, parallel_lines=2,
                                    alpha=0)
    assert w_grid == pytest.approx([0. + 0.1j, 0. + 1.j])


@pytest.mark.response
def test_mpa_sampling_2poles():
    w_grid = mpa_frequency_sampling(2, wrange=[0, 1], varpi=1,
                                    eta0=0.1, eta_rest=0.1, parallel_lines=2,
                                    alpha=1)
    assert w_grid == pytest.approx([0. + 0.1j, 1. + 0.1j, 0 + 1j, 1 + 1j])

    # print("npol=2, parallel_lines=1, w1=0+1j, w2=2+1j, alpha=0:")
    w_grid = mpa_frequency_sampling(2, wrange=[0, 2], varpi=1,
                                    eta0=0.01, eta_rest=0.1, parallel_lines=1,
                                    alpha=0)
    assert w_grid == pytest.approx([0. + 1.j, 2 / 3. + 1.j, 4 / 3. + 1.j,
                                    2. + 1.j])

    # print("npol=2, parallel_lines=2, w1=0+1j, w2=2+1j, alpha=1:")
    w_grid = mpa_frequency_sampling(2, wrange=[0, 2], varpi=1,
                                    eta0=0.01, eta_rest=0.1)
    assert w_grid == pytest.approx([0. + 0.01j, 2. + 0.1j,
                                    0. + 1.j, 2. + 1.j])


@pytest.mark.response
def test_mpa_sampling_multiple_poles():
    # print("npol=3, parallel_lines=2, w1=0+1j, w2=2+1j, alpha=1:")
    w_grid = mpa_frequency_sampling(3, wrange=[0, 2], varpi=1,
                                    eta0=0.01, eta_rest=0.1)
    assert w_grid == pytest.approx([0. + 0.01j, 1 + 0.1j, 2. + 0.1j,
                                    0. + 1.j, 1 + 1.j, 2. + 1.j])

    # print("npol=4, parallel_lines=2, w1=0+1j, w2=2+1j, alpha=1:")
    w_grid = mpa_frequency_sampling(4, wrange=[0, 2], varpi=1,
                                    eta0=0.01, eta_rest=0.1)
    assert w_grid == pytest.approx([0. + 0.01j, 0.5 + 0.1j, 1. + 0.1j,
                                    2. + 0.1j, 0. + 1.j, 0.5 + 1.j, 1. + 1.j,
                                    2. + 1.j])

    # print("npol=5, parallel_lines=2, w1=0+1j, w2=2+1j, alpha=1:")
    w_grid = mpa_frequency_sampling(5, wrange=[0, 2], varpi=1,
                                    eta0=0.01, eta_rest=0.1)
    assert w_grid == pytest.approx([0. + 0.01j, 0.25 + 0.1j, 0.5 + 0.1j,
                                    1. + 0.1j, 2. + 0.1j, 0. + 1.j, 0.25 + 1.j,
                                    0.5 + 1.j, 1. + 1.j, 2. + 1.j])

    # print("npol=6, parallel_lines=2, w1=0+1j, w2=2+1j, alpha=1:")
    w_grid = mpa_frequency_sampling(6, wrange=[0, 2], varpi=1,
                                    eta0=0.01, eta_rest=0.1)
    assert w_grid == pytest.approx([0. + 0.01j, 0.25 + 0.1j, 0.5 + 0.1j,
                                    1. + 0.1j, 1.5 + 0.1j, 2. + 0.1j, 0. + 1.j,
                                    0.25 + 1.j, 0.5 + 1.j, 1. + 1.j, 1.5 + 1.j,
                                    2. + 1.j])



import pytest
from gpaw.response.g0w0 import G0W0
import numpy as np
from gpaw.mpi import world


@pytest.mark.response
def test_gw_anisotropic(in_tmp_dir, gpw_files, gpaw_new):
    if gpaw_new and world.size > 1:
        pytest.skip('Hybrids not working in parallel with GPAW_NEW=1')
    print(gpw_files)

    gw = G0W0(gpw_files['sic_pw'],
              'gw-test',
              nbands=5,
              ecut=20,
              eta=0.2,
              frequencies={'type': 'nonlinear', 'domega0': 0.3},
              truncation=None,
              kpts=[(-0.125, 0.125, 0.125), (-0.125, -0.125, -0.125)],
              bands=(3, 5))

    e_qp = gw.calculate()['qp']

    print(e_qp)
    assert np.allclose(e_qp, [[[8.48523631, 14.83302236],
                               [7.12242815, 15.76829637]]], atol=0.001)



import pytest
from ase.build import molecule
from ase.units import Bohr
from gpaw import GPAW
from gpaw.test import gen


@pytest.mark.later
def test_S2p():
    """Compare number of electrons for different channels with 2p corehole"""
    atoms = molecule('SH2')
    atoms.center(vacuum=2)

    setup = gen('S', name='fch2p', corehole=(2, 1, 1))

    grf = 1
    calc = GPAW(mode='fd', h=0.3, spinpol=True,
                setups={'S': setup})
    atoms.calc = calc
    atoms.get_potential_energy()

    for spin, nelectrons in zip([0, 1], [8, 9]):
        n_g = calc.get_all_electron_density(spin=spin, gridrefinement=grf)
        ne = calc.density.gd.integrate(
            n_g, global_integral=False) * (Bohr / grf)**3
        assert ne == pytest.approx(nelectrons, abs=1e-5)



""" Tests extrapolation to infinite energy cutoff + block parallelization.
It takes ~10 s on one core"""

import pytest
from gpaw.response.g0w0 import G0W0
from gpaw.mpi import world
import numpy as np


@pytest.mark.response
def test_response_gw_hBN_extrapolate(in_tmp_dir, scalapack, gpw_files,
                                     gpaw_new):
    if gpaw_new and world.size > 1:
        pytest.skip('Hybrids not working in parallel with GPAW_NEW=1')
    ecuts = [20, 25, 30]
    common = dict(truncation='2D',
                  q0_correction=True,
                  kpts=[0],
                  eta=0.2,
                  bands=(3, 5),
                  nblocksmax=True)
    gw = G0W0(gpw_files['hbn_pw'],
              'gw-hBN',
              ecut=30,
              frequencies={'type': 'nonlinear',
                           'domega0': 0.1},
              ecut_extrapolation=ecuts,
              **common)

    results = gw.calculate()
    e_qp = results['qp'][0, 0]

    ev = -1.4528
    ec = 3.69469
    assert e_qp[0] == pytest.approx(ev, abs=0.01)
    assert e_qp[1] == pytest.approx(ec, abs=0.01)

    # The individual sigma matrix elements should be exactly the same
    # when running with ecut-extrapolation, and when calculating
    # with particular ecuts individually (given that nbands is not specified,
    # and it is also chosen utilizing ecut).
    for ie, ecut in enumerate(ecuts):
        gw = G0W0(gpw_files['hbn_pw'],
                  f'gw-hBN-separate-ecut{ecut}',
                  ecut=ecut,
                  frequencies={'type': 'nonlinear',
                               'omegamax': 43.2,  # We need same grid as above
                               'domega0': 0.1},
                  ecut_extrapolation=False,
                  **common)
        res = gw.calculate()
        assert np.allclose(res['dsigma_eskn'][0], results['dsigma_eskn'][ie])
        assert np.allclose(res['sigma_eskn'][0], results['sigma_eskn'][ie])



import pytest
from ase import Atoms
from ase.units import Ha

from gpaw.new.calculation import DFTCalculation
from gpaw.mpi import size
from gpaw.poisson import FDPoissonSolver
from gpaw.new.c import GPU_AWARE_MPI


@pytest.mark.gpu
@pytest.mark.serial
@pytest.mark.parametrize('dtype', [float, complex])
@pytest.mark.parametrize('gpu', [False, True])
@pytest.mark.parametrize('mode', ['pw', 'fd'])
def test_gpu(dtype, gpu, mode):
    atoms = Atoms('H2')
    atoms.positions[1, 0] = 0.75
    atoms.center(vacuum=1.0)
    if mode == 'fd':
        poisson = FDPoissonSolver()
        h = 0.17
    else:
        poisson = None
        h = None
    dft = DFTCalculation.from_parameters(
        atoms,
        dict(mode={'name': mode,
                   'force_complex_dtype': dtype == complex},
             poissonsolver=poisson,
             h=h,
             convergence={'density': 1e-8},
             parallel={'gpu': gpu},
             setups='paw'),
        log='-')
    dft.converge()
    dft.energies()
    energy = dft.results['energy'] * Ha
    if mode == 'pw':
        assert energy == pytest.approx(-16.032945, abs=1e-6)
    else:
        assert energy == pytest.approx(5.071972893296197, abs=1e-6)


@pytest.mark.gpu
@pytest.mark.skipif(size > 2, reason='Not implemented')
@pytest.mark.parametrize('gpu', [False, True])
@pytest.mark.parametrize('par', ['domain', 'kpt', 'band'])
@pytest.mark.parametrize('mode', ['pw', 'fd'])
@pytest.mark.parametrize('xc', ['LDA', 'PBE'])
def test_gpu_k(gpu, par, mode, xc):
    if gpu and size > 1 and not GPU_AWARE_MPI:
        if mode == 'fd' and par == 'domain':
            pytest.skip('Domain decomposition needs GPU-aware MPI')
        if mode == 'pw' and par == 'domain' and xc == 'PBE':
            pytest.skip('Domain decomposition needs GPU-aware MPI')
        if mode == 'fd':
            pytest.skip('???')
        if mode == 'pw' and par in ['kpt', 'band'] and xc == 'PBE':
            pytest.skip('???')

    if gpu and size > 1:
        if mode == 'fd' and par == 'band':
            pytest.skip('FAILING')

    if size == 1 and par in ['kpt', 'band']:
        pytest.skip('Not testing parallelization on single core')

    atoms = Atoms('H', pbc=True, cell=[1, 1.1, 1.1])
    if mode == 'fd':
        poisson = FDPoissonSolver()
        h = 0.09
    else:
        poisson = None
        h = None

    dft = DFTCalculation.from_parameters(
        atoms,
        dict(mode={'name': mode},
             spinpol=True,
             xc=xc,
             h=h,
             convergence={'density': 1e-8},
             kpts=(4, 1, 1),
             poissonsolver=poisson,
             parallel={'gpu': gpu,
                       par: size},
             setups='paw'),
        log='-')
    dft.converge()
    dft.energies()
    if mode == 'pw':
        dft.forces()
        dft.stress()
    energy = dft.results['energy'] * Ha
    ref = {'LDAfd': -17.685022604078714,
           'PBEfd': -17.336991943070384,
           'PBEpw': -17.304186,
           'LDApw': -17.653433}[xc + mode]
    assert energy == pytest.approx(ref, abs=1e-6)



import pytest

import numpy as np

from gpaw import GPAW
from gpaw.mpi import world
from gpaw.response import ResponseGroundStateAdapter, ResponseContext
from gpaw.response.chiks import ChiKSCalculator, SelfEnhancementCalculator
from gpaw.response.frequencies import ComplexFrequencyDescriptor
from gpaw.response.dyson import DysonSolver
from gpaw.response.susceptibility import (spectral_decomposition,
                                          read_eigenmode_lineshapes)
from gpaw.response.pair_functions import read_pair_function

from gpaw.test import findpeak
from gpaw.test.gpwfile import response_band_cutoff


@pytest.mark.kspair
@pytest.mark.response
def test_response_iron_sf_pawALDA(in_tmp_dir, gpw_files, scalapack):
    # ---------- Inputs ---------- #

    # Magnetic response calculation
    q_qc = [[0.0, 0.0, 0.0], [0.0, 0.0, 1. / 4.]]  # Two q-points along G-N
    frq_w = np.linspace(-1.0, 3.0, 161)
    ecut = 100
    eta = 0.5
    rshewmin = 1e-8

    if world.size > 1:
        nblocks = 2
    else:
        nblocks = 1

    # ---------- Script ---------- #

    context = ResponseContext(txt='iron_susceptibility.txt')
    calc = GPAW(gpw_files['fe_pw'], parallel=dict(domain=1))
    nbands = response_band_cutoff['fe_pw']
    gs = ResponseGroundStateAdapter(calc)

    calc_args = (gs,)
    calc_kwargs = dict(context=context,
                       nbands=nbands,
                       ecut=ecut,
                       gammacentered=True,
                       bandsummation='double',
                       nblocks=nblocks)
    chiks_calc = ChiKSCalculator(*calc_args, **calc_kwargs)
    xi_calc = SelfEnhancementCalculator(*calc_args,
                                        rshewmin=rshewmin,
                                        **calc_kwargs)
    dyson_solver = DysonSolver(context)

    for q, q_c in enumerate(q_qc):
        # Calculate χ_KS^+- and Ξ^++
        zd = ComplexFrequencyDescriptor.from_array(frq_w + 1j * eta)
        chiks = chiks_calc.calculate('+-', q_c, zd)
        xi = xi_calc.calculate('+-', q_c, zd)

        # Distribute frequencies and invert dyson equation
        chiks = chiks.copy_with_global_frequency_distribution()
        xi = xi.copy_with_global_frequency_distribution()
        chi = dyson_solver(chiks, xi)

        # plot_inverse_enhancement(xi)

        # Write macroscopic component and acoustic magnon mode
        chi.write_macroscopic_component(f'iron_chiM_q{q}.csv')
        Amaj, _ = spectral_decomposition(chi)
        Amaj.write_eigenmode_lineshapes(f'iron_Amaj_q{q}.csv')

    context.write_timer()

    # plot_magnons()

    refs_q = [
        # (wpeak, Ipeak, Apeak)
        (0.001, 0.476, 3.104),
        (0.165, 0.428, 2.816),
    ]
    for q, refs in enumerate(refs_q):
        w_w, chiM_w, a_w = extract_data(q)
        wpeak1, Ipeak = findpeak(w_w, -chiM_w.imag / np.pi)
        wpeak2, Apeak = findpeak(w_w, a_w)

        assert wpeak1 == pytest.approx(wpeak2, abs=0.002)  # eV
        assert wpeak1 == pytest.approx(refs[0], abs=0.01)  # eV
        assert Ipeak == pytest.approx(refs[1], abs=0.01)  # a.u.
        assert Apeak == pytest.approx(refs[2], abs=0.05)  # a.u.


def extract_data(q):
    w1_w, chiM_w = read_pair_function(f'iron_chiM_q{q}.csv')
    w2_w, a_wm = read_eigenmode_lineshapes(f'iron_Amaj_q{q}.csv')
    assert np.allclose(w1_w, w2_w)
    return w1_w, chiM_w, a_wm[:, 0]


def plot_inverse_enhancement(xi):
    import matplotlib.pyplot as plt
    from ase.units import Ha
    invenh_mywM, _ = np.linalg.eig(xi.array)
    invenh_wM = xi.blocks1d.all_gather(invenh_mywM)
    omega_w = xi.zd.omega_w * Ha

    for M in range(invenh_wM.shape[1]):
        plt.subplot(1, 2, 1)
        plt.scatter(omega_w, invenh_wM[:, M].real)
        plt.subplot(1, 2, 2)
        plt.scatter(omega_w, invenh_wM[:, M].imag)
    plt.subplot(1, 2, 1)
    plt.axhline(1., c='0.5')
    if world.rank == 0:
        plt.show()


def plot_magnons():
    import matplotlib.pyplot as plt
    for q in range(2):
        w_w, chiM_w, a_w = extract_data(q)
        plt.subplot(1, 2, 1)
        plt.plot(w_w, - chiM_w.imag / np.pi, label=f'{q}')
        plt.subplot(1, 2, 2)
        plt.plot(w_w, a_w, label=f'{q}')
        plt.legend(title='q')
    if world.rank == 0:
        plt.show()



r""" Class which builds a weight function for each atom
    in a molecule using simple atom centred gaussians.

    This is the Hirshfeld scheme:
               A
     A        p (r)
    w (r) = ---------
             mol
             \   n
             /  p (r)
             --
             n

"""

import numpy as np
from math import pi
from ase.units import Bohr

# Cut-offs: [Ang]
Rc = {'Fe': 3.762,
      'O': 3.762,
      # 'H' : 3.762
      }

# Gauss-width: [Ang]
mu = {'Fe': 1.0,
      'O': 0.6,
      # 'H' : 0.6
      }


class WeightFunc:

    def __init__(self, gd, atoms, indexes, Rc=Rc, mu=mu):
        """ Given a grid-descriptor, atoms object and an index list
            construct a weight function defined by:

        """

        self.gd = gd  # Grid descriptor
        self.atoms = atoms
        self.indexes = indexes

        # Construct Rc dict
        new = {}
        for a in self.atoms:
            if a.symbol in Rc:
                new[a.symbol] = Rc[a.symbol]
            else:
                new[a.symbol] = 3.762

        self.Rc = new

        # Construct mu dict
        new_mu = {}
        for a in self.atoms:
            if mu:
                if a.symbol in mu:
                    new_mu[a.symbol] = mu[a.symbol]
            else:
                new_mu[a.symbol] = 0.85

        # Larger atoms may need a bit more width?
        self.mu = new_mu

    def truncated_gaussian(self, dis, mu, Rc):
        # Given mu and Rc construct Gaussian.
        # Gauss. is truncated at Rc

        check = abs(dis) <= Rc / Bohr
        # Make gaussian
        gauss = 1.0 / (mu * np.sqrt(2.0 * pi)) * \
            np.exp(- dis ** 2 / (2.0 * mu))
        # Apply cut-off and send
        return (gauss * check)

    def get_distance_vectors(self, pos):
        # Given atom position [Bohr], grab distances to all
        # grid-points (gpts) - employ MIC where appropriate.

        # Scaled positions of gpts on some cpu, relative to all
        s_G = (np.indices(self.gd.n_c, float).T +
               self.gd.beg_c) / self.gd.N_c
        # print self.gd.n_c, self.gd.beg_c
        # Subtract scaled distance from atom to box boundary
        s_G -= np.linalg.solve(self.gd.cell_cv.T, pos)
        # MIC
        s_G -= self.gd.pbc_c * (2 * s_G).astype(int)
        # Apparently doing this twice works better...
        s_G -= self.gd.pbc_c * (2 * s_G).astype(int)
        # x,y,z distances
        xyz = np.dot(s_G, self.gd.cell_cv).T.copy()
        #
        return np.sqrt((xyz ** 2).sum(axis=0))

    def construct_total_density(self, atoms):
        # Add to empty grid
        empty = self.gd.zeros()

        for atom in atoms:
            charge = atom.number
            symbol = atom.symbol

            pos = atom.position / Bohr

            dis = self.get_distance_vectors(pos)

            empty += charge * self.truncated_gaussian(
                dis, self.mu[symbol], self.Rc[symbol])
        #
        return empty

    def construct_weight_function(self):
        # Grab atomic / molecular density
        dens_n = self.construct_total_density(
            self.atoms[self.indexes])
        # Grab total density
        dens = self.construct_total_density(self.atoms)
        # Check zero elements
        check = dens == 0
        # Add constant to zeros ...
        dens += check * 1.0
        # make and send
        return dens_n / dens


class WignerSeitz:
    """ Construct weight function based on Wigner-Seitz

               | 1 if (r-R) < (r-R')
        w(r) = |
               | 0 if (r-R) > (r-R')

        for atom A at pos R

    """

    def __init__(self, gd, atoms, index):
        #
        self.gd = gd
        self.atoms = atoms
        self.index = index

    def get_distance_vectors(self, pos):
        #
        # Given atom position [Bohr], grab distances to all
        # grid-points (gpts) - employ MIC where appropriate.

        # Scaled positions of gpts on some cpu, relative to all
        s_G = (np.indices(self.gd.n_c, float).T +
               self.gd.beg_c) / self.gd.N_c
        #
        # Subtract scaled distance from atom to box boundary
        s_G -= np.linalg.solve(self.gd.cell_cv.T, pos)
        # MIC
        s_G -= self.gd.pbc_c * (2 * s_G).astype(int)
        # Apparently doing this twice works better...
        s_G -= self.gd.pbc_c * (2 * s_G).astype(int)
        # x,y,z distances
        xyz = np.dot(s_G, self.gd.cell_cv).T.copy()
        #
        return np.sqrt((xyz ** 2).sum(axis=0))

    def construct_weight_function(self):
        # Grab distances of A to all gpts
        pos = self.atoms[self.index].position / Bohr
        dis_n = self.get_distance_vectors(pos)
        #
        empty = self.gd.zeros()
        norm = self.gd.zeros()
        #
        empty += 1
        norm += 1
        #
        for i, atom in enumerate(self.atoms):
            #
            if i == self.index:
                continue
            else:
                #
                pos = atom.position / Bohr
                dis_b = self.get_distance_vectors(pos)
                #
                check = dis_n <= dis_b  #
                n_c = dis_n == dis_b  #
                # 0 if longer, 1 if shorter, 1/no.(Ra=Ra') if same
                empty *= check
                norm += n_c * 1.0

        return empty / norm



import pytest
import numpy as np

from ase import Atoms
from ase.parallel import parprint
from ase.units import Hartree

import gpaw.mpi as mpi
from gpaw import GPAW
from gpaw.lrtddft.kssingle import KSSingles


@pytest.mark.lrtddft
def test_lrtddft_kssingles_Be(in_tmp_dir):
    Be = Atoms('Be')
    Be.center(vacuum=4)
    if 1:
        # introduce a sligth non-orthgonality
        cell = Be.get_cell()
        cell[1] += 0.001 * cell[0]
        Be.set_cell(cell)

    txt = None
    eigensolver = None

    # modes = ['lcao', 'fd']
    modes = ['fd']

    for mode in modes:
        energy = {}
        osz = {}
        for pbc in [False, True]:
            Be.set_pbc(pbc)
            if pbc:
                name = 'periodic'
                calc = GPAW(h=0.25,
                            nbands=4,
                            kpts=(1, 2, 2),
                            mode=mode,
                            poissonsolver={'name': 'fd'},
                            symmetry='off',
                            eigensolver=eigensolver,
                            txt=txt)
            else:
                name = 'zero_bc'
                calc = GPAW(h=0.25, nbands=4, mode=mode,
                            poissonsolver={'name': 'fd'},
                            eigensolver=eigensolver, txt=txt)
            Be.calc = calc
            Be.get_potential_energy()

            kss = KSSingles(restrict={'eps': 0.9})
            kss.calculate(Be)
            # all s->p transitions at the same energy [Ha] and
            # oscillator_strength
            for ks in kss:
                assert ks.get_energy() == pytest.approx(kss[0].get_energy(),
                                                        abs=5.e-3)
                assert ks.get_oscillator_strength()[0] == pytest.approx(
                    kss[0].get_oscillator_strength()[0], abs=5.e-3)
                assert ks.get_oscillator_strength()[0] == pytest.approx(
                    ks.get_oscillator_strength()[1:].sum() / 3, abs=1.e-15)
                for c in range(3):
                    assert ks.get_oscillator_strength()[1 + c] == (
                        pytest.approx(ks.get_dipole_tensor()[c, c],
                                      abs=1.e-15))
            energy[name] = np.array(
                [ks.get_energy() * Hartree for ks in kss]).mean()
            osz[name] = np.array(
                [ks.get_oscillator_strength()[0] for ks in kss]).sum()

            parprint(name + ':')
            parprint(kss)

            # I/O
            fname = 'kss_' + name + '.dat'
            kss.write(fname)
            mpi.world.barrier()
            kss = KSSingles.read(fname)
            kss1 = KSSingles.read(fname, restrict={'jend': 1})
            assert len(kss1) == calc.wfs.kd.nibzkpts * calc.wfs.nspins

        # periodic and non-periodic should be roughly equal
        assert energy['zero_bc'] == pytest.approx(energy['periodic'],
                                                  abs=5.e-2)
        assert osz['zero_bc'] == pytest.approx(osz['periodic'], abs=2.e-2)



"""Test the reading of wave functions as file references."""
import numpy as np
from gpaw import GPAW
from gpaw.mpi import rank, world


def test_fileio_file_reference(in_tmp_dir, gpw_files):
    # load restart from gpw
    calc = GPAW(gpw_files['na3_fd_kp_restart'])
    wf0 = calc.get_pseudo_wave_function(2, 1, 1)

    # Now read with a single process
    comm = world.new_communicator(np.array((rank,)))
    calc = GPAW(gpw_files['na3_fd_kp_restart'], communicator=comm)
    wf1 = calc.get_pseudo_wave_function(2, 1, 1)

    # compare wf restarts match
    diff = np.abs(wf0 - wf1)
    assert np.all(diff < 1e-12)



import numpy as np
import pytest

from gpaw.mpi import world
from gpaw.new.ase_interface import GPAW
from gpaw.utilities import pack_density


@pytest.mark.soc
@pytest.mark.skipif(world.size > 1, reason='Gamma-point calculation.')
def test_kinetic_energy(gpw_files):
    # Test that we calculate the kinetic energy correctly in noncollinear mode.
    # We are mostly concerned with the contributions from the complex parts of
    # the atomic density matrix and Hamiltonian (spin-orbit).

    # Thallium atom in a simulation box, we use a heavy atom with large SOC.
    calc = GPAW(gpw_files['Tl_box_pw'])
    state = calc.dft.state
    setup = calc.dft.setups[0]

    # Kinetic energy calculated from sum of bands (the standard way):
    Ekin1 = state.ibzwfs.energies['band'] + state.potential.energies['kinetic']

    # Kinetic energy calculated directly from second-order derivative:
    wfs = state.ibzwfs.wfs_qs[0][0]

    psit = wfs.psit_nX
    occ_n = wfs.occ_n

    psit_nsG = psit.data
    G_plus_k_Gv = psit.desc.G_plus_k_Gv
    ucvol = psit.desc.volume

    laplacian_on_psit_nsv = np.abs(psit_nsG)**2 @ G_plus_k_Gv**2
    laplacian_on_psit_n = - np.sum(np.sum(laplacian_on_psit_nsv, axis=1),
                                   axis=1) * ucvol
    Ekin_pseudo = -0.5 * (laplacian_on_psit_n @ occ_n)

    D_p = pack_density(state.density.D_asii[0][0].real)
    Ekin_PAW = setup.K_p @ D_p + setup.Kc

    Ekin2 = Ekin_pseudo + Ekin_PAW

    # We should get the same value for the kinetic
    # energy irrespective of the method used.
    assert Ekin1 == pytest.approx(Ekin2, abs=1e-5)



import numpy as np
from time import time

from gpaw.response.mpa_sampling import mpa_frequency_sampling
from gpaw.response.mpa_interpolation import RESolver, fit_residue

from .mpa_interpolation_scalar import mpa_RE_solver, mpa_R_fit, Xeval


def test_pole_reduction(in_tmp_dir):
    npols, npol_fit = 1, 2
    rng = np.random.default_rng(seed=42)
    Omega_p = rng.random(npols) * 0.05 + 5.5 - 0.01j
    residues_p = rng.random(npols)

    omega_w = mpa_frequency_sampling(
        npoles=npol_fit,
        wrange=[0, 0.8],
        varpi=1,
        eta0=0.1,
        eta_rest=0.1,
        parallel_lines=2,
    )

    X_w = Xeval(Omega_p, residues_p, omega_w)
    R_p, E_p, _, _ = mpa_RE_solver(npol_fit, omega_w, X_w[0, 0])
    E_pGG, R_pGG = RESolver(omega_w).solve(X_w.reshape((-1, 1, 1)))

    E_GGp = E_pGG.transpose(1, 2, 0)
    R_GGp = R_pGG.transpose(1, 2, 0)

    assert np.allclose(E_GGp[0, 0, :], E_p, rtol=1e-4, atol=1e-6)
    assert np.allclose(R_GGp[0, 0], R_p, rtol=1e-4, atol=1e-6)

    if 0:
        from matplotlib import pyplot as plt
        w_x = np.linspace(0, 1, 1000)
        X_x = Xeval(Omega_p, residues_p, w_x)
        fit_x = Xeval(E_p, R_p, w_x)
        fit_vec_x = Xeval(E_GGp[0, 0, :], R_GGp[0, 0, :], w_x)

        plt.plot(w_x, X_x[0, 0, :].real, 'r')
        plt.plot(w_x, X_x[0, 0, :].imag, 'g')

        plt.plot(w_x, fit_vec_x[0, 0, :].real, 'r--')
        plt.plot(w_x, fit_vec_x[0, 0, :].imag, 'g--')

        plt.plot(w_x, fit_x[0, 0, :].real, 'k:')
        plt.plot(w_x, fit_x[0, 0, :].imag, 'k:')
        plt.show()


def test_residue_fit_1pole(in_tmp_dir):
    npols, npr, w, x, E = (
        1,
        1,
        np.array([0.0 + 0.0j, 0.0 + 1.0j]),
        np.array([0.13034393 - 0.40439649j, 0.36642415 - 0.67018998j]),
        np.array([1.654376 - 0.25302243j]),
    )

    R = mpa_R_fit(npols, npr, w, x, E)
    R_vec = fit_residue(npr_GG=np.array([[npols]]),
                        omega_w=w,
                        X_wGG=x.reshape((-1, 1, 1)),
                        E_pGG=E.reshape((-1, 1, 1)))
    assert np.allclose(R, R_vec)


def test_residue_fit(in_tmp_dir):
    npols, npr, w, x, E = (
        2,
        2,
        np.array([0.0 + 0.0j, 0.0 + 1.0j, 2.33 + 0.0j, 2.33 + 1.0j]),
        np.array(
            [
                0.13034393 - 0.40439649j,
                0.36642415 - 0.67018998j,
                0.77096523 + 0.85578656j,
                -0.38002124 - 0.20843867j,
            ]
        ),
        np.array([1.654376 - 0.25302243j, 2.4306366 - 0.15733093j]),
    )

    R = mpa_R_fit(npols, npr, w, x, E)
    R_vec = fit_residue(npr_GG=np.array([[npols]]),
                        omega_w=w,
                        X_wGG=x.reshape((-1, 1, 1)),
                        E_pGG=E.reshape((-1, 1, 1)))
    assert np.allclose(R, R_vec[:, 0, 0])


def test_ppa(in_tmp_dir):
    nG = 120
    rng = np.random.default_rng(seed=42)
    X_wGG = (2 * (rng.random((2, nG, nG)) - 0.5) + 1j *
             (rng.random((2, nG, nG)) - 0.5) * 2)
    omega_w = np.array([0, 2j])

    start = time()
    E_GG, R_GG = RESolver(omega_w).solve(X_wGG)
    stop = time()
    fail = False
    for i in range(nG):
        for j in range(nG):
            R, E, MPres, PPcond_rate = mpa_RE_solver(
                npols=1, w=omega_w, x=X_wGG[:, i, j])
            assert np.allclose(E, E_GG[i, j])
            if not np.allclose(R, R_GG[i, j]):
                fail = True
    assert not fail
    vectorized_time = stop - start
    print('\nVectorized', vectorized_time)
    start = time()
    for i in range(nG):
        for j in range(nG):
            R, E, _, _ = mpa_RE_solver(
                npols=1, w=omega_w, x=X_wGG[:, i, j])
    stop = time()
    serial_time = stop - start
    print('Not vectorized', serial_time)
    print('speedup', serial_time / vectorized_time)


def test_mpa(in_tmp_dir):
    nG = 8
    omega_w = np.array([0, 1j, 2. + 0.01j, 2. + 1j])
    nw = len(omega_w)
    rng = np.random.default_rng(seed=42)
    X_wGG = (2 * (rng.random((nw, nG, nG)) - 0.5) + 1j *
             (rng.random((nw, nG, nG)) - 0.5) * 2)

    E_pGG, R_pGG = RESolver(omega_w).solve(X_wGG)
    E_GGp = E_pGG.transpose(1, 2, 0)
    R_GGp = R_pGG.transpose(1, 2, 0)

    for i in range(nG):
        for j in range(nG):
            R_p, E_p, _, _ = mpa_RE_solver(
                npols=nw // 2, w=omega_w, x=X_wGG[:, i, j])
            ind = np.argsort(E_p.real)
            E_p = E_p[ind]
            R_p = R_p[ind]
            assert np.allclose(E_GGp[i, j, :], E_p, rtol=1e-5, atol=1e-6)
            assert np.allclose(R_GGp[i, j, :], R_p, rtol=1e-5, atol=1e-6)



import numpy as np
import pytest

from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.mpi import world


@pytest.mark.rttddft
def test_lcaotddft_fxc_rpa(gpw_files, in_tmp_dir):
    # Time-propagation calculation with fxc
    td_calc = LCAOTDDFT(gpw_files['na2_tddft_poisson'], fxc='RPA',
                        txt='td.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    td_calc.absorption_kick(np.ones(3) * 1e-5)
    td_calc.propagate(20, 3)
    world.barrier()

    # Check dipole moment file
    data = np.loadtxt('dm.dat')[:, 2:].ravel()
    if 0:
        from gpaw.test import print_reference
        print_reference(data, 'ref', '%.12le')

    ref = [-9.383700894739e-16,
           -9.338586948130e-16,
           2.131582675483e-14,
           8.679923327633e-15,
           7.529517689096e-15,
           2.074867751820e-14,
           1.960742702185e-05,
           1.960742702128e-05,
           1.804030746475e-05,
           3.761997205571e-05,
           3.761997205562e-05,
           3.596681520068e-05,
           5.257367158160e-05,
           5.257367158104e-05,
           5.366663490365e-05]

    tol = 1e-8
    assert data == pytest.approx(ref, abs=tol)



import numpy as np
import pytest
from gpaw.spinorbit import get_L_vlmm


class QuantumState:
    """Quantum State, representing real Y_lm or complex Y_l^m"""
    def __init__(self, l: int, m: int, factor: complex = 1.0):
        assert type(l) is int
        assert type(m) is int
        self.l = l
        self.m = m
        self.factor = factor


def real_to_complex_quantum_state(state: QuantumState):
    """represent the real Y_lm in complex Y_l^m"""
    if state.m == 0:
        return [QuantumState(state.l, state.m, 1.0)]
    elif state.m > 0:
        return [
            QuantumState(state.l, -1 * state.m,
                         1 / np.sqrt(2.0) * state.factor),
            QuantumState(state.l, state.m,
                         (-1.0) ** state.m / np.sqrt(2.0) * state.factor)]
    else:  # state.m < 0:
        return [
            QuantumState(state.l, state.m, 1j / np.sqrt(2.0) * state.factor),
            QuantumState(
                state.l,
                -1 * state.m,
                -1j * (-1.0) ** state.m / np.sqrt(2.0) * state.factor,
            ),
        ]


def matrix_element(bra: QuantumState, ket: QuantumState, operator):
    """<Y_lm' | O | Y_lm >"""
    assert bra.l == ket.l
    result = 0.0
    kets_complex = real_to_complex_quantum_state(ket)
    bras_complex = real_to_complex_quantum_state(bra)
    for b in bras_complex:
        for k in kets_complex:
            operator_factor, newk = operator(k)
            if newk.m != b.m:
                continue
            result += np.conj(b.factor) * newk.factor * operator_factor
    return result


def Lz(ket: QuantumState):
    """L_z acting on complex Y_l^m"""
    return ket.m, QuantumState(ket.l, ket.m, ket.factor)


def Lp(ket: QuantumState):
    """L_+ acting on complex Y_l^m"""
    if ket.m < ket.l:
        result = np.sqrt(ket.l * (ket.l + 1) - ket.m * (ket.m + 1))
    else:
        result = 0
    return result, QuantumState(ket.l, ket.m + 1, ket.factor)


def Lm(ket: QuantumState):
    """L_- acting on complex Y_l^m"""
    if ket.m > -ket.l:
        result = np.sqrt(ket.l * (ket.l + 1) - ket.m * (ket.m - 1))
    else:
        result = 0
    return result, QuantumState(ket.l, ket.m - 1, ket.factor)


def test_lvlmm():

    L_vlmm = get_L_vlmm()

    for l in [0, 1, 2, 3]:
        # print(f"Doing {'spdf'[l]}:")
        result_lz = np.zeros((2 * l + 1, 2 * l + 1), dtype=complex)
        result_lp = np.zeros((2 * l + 1, 2 * l + 1), dtype=complex)
        result_lm = np.zeros((2 * l + 1, 2 * l + 1), dtype=complex)
        for mp in range(-l, l + 1):
            bra = QuantumState(l, mp)  # real Y_lm
            for m in range(-l, l + 1):
                ket = QuantumState(l, m)  # real Y_lm
                result_lz[mp + l, m + l] = matrix_element(bra, ket, Lz)
                result_lp[mp + l, m + l] = matrix_element(bra, ket, Lp)
                result_lm[mp + l, m + l] = matrix_element(bra, ket, Lm)

        result_lx = (result_lp + result_lm) / 2.0
        result_ly = (result_lp - result_lm) / 2j

        L_vlmm[0][l] == pytest.approx(result_lx)
        L_vlmm[1][l] == pytest.approx(result_ly)
        L_vlmm[2][l] == pytest.approx(result_lz)



"""
Optimization of orbitals
among occupied and a few virtual states
represented on a grid or with plane waves
in order to calculate and excited state

arXiv:2102.06542 [physics.comp-ph]
"""

from gpaw.directmin.tools import get_n_occ, get_indices, expm_ed, \
    sort_orbitals_according_to_occ
from gpaw.directmin.sd_etdm import LSR1P
from gpaw.directmin.ls_etdm import MaxStep
from gpaw.directmin.derivatives import get_approx_analytical_hessian
from ase.units import Hartree
import numpy as np
import time


class ETDMInnerLoop:

    def __init__(self, odd_pot, wfs, nstates='all', maxiter=100,
                 maxstepxst=0.2, g_tol=5.0e-4, useprec=False, momevery=10):

        self.odd_pot = odd_pot
        self.n_kps = wfs.kd.nibzkpts
        self.g_tol = g_tol / Hartree
        self.dtype = wfs.dtype
        self.get_en_and_grad_iters = 0
        self.precond = {}
        self.max_iter_line_search = 6
        self.n_counter = maxiter
        self.maxstep = maxstepxst
        self.eg_count = 0
        self.total_eg_count = 0
        self.run_count = 0
        self.U_k = {}
        self.Unew_k = {}
        self.e_total = 0.0
        self.n_occ = {}
        self.useprec = useprec
        for kpt in wfs.kpt_u:
            k = self.kpointval(kpt)
            if nstates == 'all':
                self.n_occ[k] = wfs.bd.nbands
            elif nstates == 'occupied':
                self.n_occ[k] = get_n_occ(kpt)[0]
            else:
                raise NotImplementedError
            self.U_k[k] = np.eye(self.n_occ[k], dtype=self.dtype)
            self.Unew_k[k] = np.eye(self.n_occ[k], dtype=self.dtype)
        self.momcounter = 0
        self.momevery = momevery
        self.restart = False
        self.eks = 0.0
        self.esic = 0.0
        self.kappa = 0.0

    def update_ks_energy(self, wfs, dens, ham):
        wfs.timer.start('Update Kohn-Sham energy')
        # calc projectors
        for kpt in wfs.kpt_u:
            wfs.pt.integrate(kpt.psit_nG, kpt.P_ani, kpt.q)

        dens.update(wfs)
        ham.update(dens, wfs, False)
        wfs.timer.stop('Update Kohn-Sham energy')
        return ham.get_energy(0.0, wfs, False)

    def get_energy_and_gradients(self, a_k, wfs, dens, ham):
        """
        Energy E = E[A]. Gradients G_ij[A] = dE/dA_ij
        Returns E[A] and G[A] at psi = exp(A).T kpt.psi
        :param a_k: A
        :return:
        """

        g_k = {}
        self.e_total = 0.0
        self.esic = 0.0
        self.kappa = 0.0
        for k, kpt in enumerate(wfs.kpt_u):
            n_occ = self.n_occ[k]
            if n_occ == 0:
                g_k[k] = np.zeros_like(a_k[k])

        evecs, evals = self.rotate_wavefunctions(wfs, a_k)

        self.eks = self.update_ks_energy(wfs, dens, ham)

        for k, kpt in enumerate(wfs.kpt_u):
            wfs.timer.start('Energy and gradients')
            g_k[k], esic, kappa1 = \
                self.odd_pot.get_energy_and_gradients_inner_loop(
                    wfs, kpt, a_k[k], evals[k], evecs[k], ham=ham,
                    exstate=True)
            wfs.timer.stop('Energy and gradients')
            if kappa1 > self.kappa:
                self.kappa = kappa1
            self.esic += esic

        self.check_mom(wfs, dens)
        self.e_total = self.eks + self.esic

        self.kappa = wfs.kd.comm.max_scalar(self.kappa)
        self.eg_count += 1
        self.total_eg_count += 1

        return self.e_total, g_k

    def rotate_wavefunctions(self, wfs, a_k):
        evecs = {}
        evals = {}
        for k, kpt in enumerate(wfs.kpt_u):
            n_occ = self.n_occ[k]
            if n_occ == 0:
                continue
            wfs.timer.start('Unitary matrix')
            u_mat, evecs[k], evals[k] = expm_ed(a_k[k], evalevec=True)
            wfs.timer.stop('Unitary matrix')
            self.Unew_k[k] = u_mat.copy()
            kpt.psit_nG[:n_occ] = \
                np.tensordot(u_mat.T, self.psit_knG[k][:n_occ], axes=1)
            # calc projectors
            wfs.pt.integrate(kpt.psit_nG, kpt.P_ani, kpt.q)

        return evecs, evals

    def evaluate_phi_and_der_phi(self, a_k, p_k, alpha,
                                 wfs, dens, ham,
                                 phi=None, g_k=None):
        """
        phi = f(x_k + alpha_k*p_k)
        der_phi = grad f(x_k + alpha_k*p_k) cdot p_k
        :return:  phi, der_phi, grad f(x_k + alpha_k*p_k)
        """
        if phi is None or g_k is None:
            x_k = {k: a_k[k] + alpha * p_k[k] for k in a_k.keys()}
            phi, g_k = \
                self.get_energy_and_gradients(x_k, wfs, dens, ham)
            del x_k
        else:
            pass

        der_phi = 0.0
        for k in p_k.keys():
            il1 = get_indices(p_k[k].shape[0])
            der_phi += np.dot(g_k[k][il1].conj(), p_k[k][il1]).real

        der_phi = wfs.kd.comm.sum_scalar(der_phi)

        return phi, der_phi, g_k

    def get_search_direction(self, a_k, g_k, wfs):

        # structure of vector is
        # (x_1_up, x_2_up,..,y_1_up, y_2_up,..,
        #  x_1_down, x_2_down,..,y_1_down, y_2_down,.. )

        a = {}
        g = {}

        for k in a_k.keys():
            il1 = get_indices(a_k[k].shape[0])
            a[k] = a_k[k][il1]
            g[k] = g_k[k][il1]

        p = self.sd.update_data(wfs, a, g, precond=self.precond, mode='lcao')
        del a, g

        p_k = {}
        for k in p.keys():
            p_k[k] = np.zeros_like(a_k[k])
            il1 = get_indices(a_k[k].shape[0])
            p_k[k][il1] = p[k]
            # make it skew-hermitian
            ind_l = np.tril_indices(p_k[k].shape[0], -1)
            p_k[k][(ind_l[1], ind_l[0])] = -p_k[k][ind_l].conj()
        del p

        return p_k

    def run(self, wfs, dens, log, outer_counter=0, ham=None):

        log = log
        self.run_count += 1
        self.counter = 0
        self.eg_count = 0
        self.momcounter = 1
        self.converged = False
        # initial things
        self.psit_knG = {}
        for kpt in wfs.kpt_u:
            k = self.kpointval(kpt)
            n_occ = self.n_occ[k]
            self.psit_knG[k] = np.tensordot(
                self.U_k[k].T, kpt.psit_nG[:n_occ], axes=1)

        a_k = {}
        for kpt in wfs.kpt_u:
            k = self.kpointval(kpt)
            d = self.n_occ[k]
            a_k[k] = np.zeros(shape=(d, d), dtype=self.dtype)

        self.sd = LSR1P(memory=50)
        self.ls = MaxStep(self.evaluate_phi_and_der_phi, max_step=self.maxstep)

        threelasten = []
        # get initial energy and gradients
        self.e_total, g_k = self.get_energy_and_gradients(a_k, wfs, dens, ham)
        threelasten.append(self.e_total)
        g_max = g_max_norm(g_k, wfs)
        if g_max < self.g_tol:
            self.converged = True
            for kpt in wfs.kpt_u:
                k = self.kpointval(kpt)
                n_occ = self.n_occ[k]
                kpt.psit_nG[:n_occ] = np.tensordot(
                    self.U_k[k].conj(), self.psit_knG[k], axes=1)
                # calc projectors
                wfs.pt.integrate(kpt.psit_nG, kpt.P_ani, kpt.q)

                self.U_k[k] = self.U_k[k] @ self.Unew_k[k]
            if outer_counter is None:
                return self.e_total, self.counter
            else:
                return self.e_total, outer_counter

        if self.restart:
            del self.psit_knG
            return 0.0, 0

        # stuff which are needed for minim.
        phi_0 = self.e_total
        phi_old = None
        der_phi_old = None
        phi_old_2 = None
        der_phi_old_2 = None

        outer_counter += 1
        if log is not None:
            log_f(log, self.counter, self.kappa, self.eks, self.esic,
                  outer_counter, g_max)

        alpha = 1.0
        not_converged = True
        while not_converged:
            self.precond = self.update_preconditioning(wfs, self.useprec)

            # calculate search direction fot current As and Gs
            p_k = self.get_search_direction(a_k, g_k, wfs)

            # calculate derivative along the search direction
            phi_0, der_phi_0, g_k = \
                self.evaluate_phi_and_der_phi(
                    a_k, p_k, 0.0, wfs, dens, ham=ham, phi=phi_0, g_k=g_k)
            if self.counter > 1:
                phi_old = phi_0
                der_phi_old = der_phi_0

            # choose optimal step length along the search direction
            # also get energy and gradients for optimal step
            alpha, phi_0, der_phi_0, g_k = \
                self.ls.step_length_update(
                    a_k, p_k, wfs, dens, ham, mode='lcao',
                    phi_0=phi_0, der_phi_0=der_phi_0,
                    phi_old=phi_old_2, der_phi_old=der_phi_old_2,
                    alpha_max=3.0, alpha_old=alpha, kpdescr=wfs.kd)

            # broadcast data is gd.comm > 1
            if wfs.gd.comm.size > 1:
                alpha_phi_der_phi = np.array([alpha, phi_0, der_phi_0])
                wfs.gd.comm.broadcast(alpha_phi_der_phi, 0)
                alpha = alpha_phi_der_phi[0]
                phi_0 = alpha_phi_der_phi[1]
                der_phi_0 = alpha_phi_der_phi[2]
                for kpt in wfs.kpt_u:
                    k = self.kpointval(kpt)
                    if self.n_occ[k] == 0:
                        continue
                    wfs.gd.comm.broadcast(g_k[k], 0)

            phi_old_2 = phi_old
            der_phi_old_2 = der_phi_old

            if self.restart:
                if log is not None:
                    log('MOM has detected variational collapse, '
                        'occupied orbitals have changed')
                break

            if alpha > 1.0e-10:
                # calculate new matrices at optimal step lenght
                a_k = {k: a_k[k] + alpha * p_k[k] for k in a_k.keys()}
                g_max = g_max_norm(g_k, wfs)

                # output
                self.counter += 1
                if outer_counter is not None:
                    outer_counter += 1
                if log is not None:
                    log_f(
                        log, self.counter, self.kappa, self.eks, self.esic,
                        outer_counter, g_max)

                not_converged = \
                    g_max > self.g_tol and \
                    self.counter < self.n_counter
                if g_max <= self.g_tol:
                    self.converged = True
            else:
                break

        if log is not None:
            log('INNER LOOP FINISHED.\n')
            log('Total number of e/g calls:' + str(self.eg_count))

        if not self.restart:
            for kpt in wfs.kpt_u:
                k = self.kpointval(kpt)
                n_occ = self.n_occ[k]
                kpt.psit_nG[:n_occ] = np.tensordot(self.U_k[k].conj(),
                                                   self.psit_knG[k],
                                                   axes=1)
                # calc projectors
                wfs.pt.integrate(kpt.psit_nG, kpt.P_ani, kpt.q)
                self.U_k[k] = self.U_k[k] @ self.Unew_k[k]

        if outer_counter is None:
            return self.e_total, self.counter
        else:
            return self.e_total, outer_counter

    def update_preconditioning(self, wfs, use_prec):
        counter = 30
        if use_prec:
            if self.counter % counter == 0:
                for kpt in wfs.kpt_u:
                    k = self.kpointval(kpt)
                    hess = get_approx_analytical_hessian(kpt, self.dtype)
                    if self.dtype is float:
                        self.precond[k] = np.zeros_like(hess)
                        for i in range(hess.shape[0]):
                            if abs(hess[i]) < 1.0e-4:
                                self.precond[k][i] = 1.0
                            else:
                                self.precond[k][i] = 1.0 / hess[i].real
                    else:
                        self.precond[k] = np.zeros_like(hess)
                        for i in range(hess.shape[0]):
                            if abs(hess[i]) < 1.0e-4:
                                self.precond[k][i] = 1.0 + 1.0j
                            else:
                                self.precond[k][i] = \
                                    1.0 / hess[i].real + 1.0j / hess[i].imag
                return self.precond
            else:
                return self.precond
        else:
            return None

    def check_mom(self, wfs, dens):
        if self.momcounter % self.momevery == 0:
            occ_name = getattr(wfs.occupations, "name", None)
            if occ_name == 'mom':
                wfs.calculate_occupation_numbers(dens.fixed)
                self.restart = sort_orbitals_according_to_occ(
                    wfs, update_mom=True, update_eps=False)
        self.momcounter += 1

    def kpointval(self, kpt):
        return self.n_kps * kpt.s + kpt.q


def log_f(log, niter, kappa, eks, esic, outer_counter=None, g_max=np.inf):

    t = time.localtime()

    if niter == 0:
        header0 = '\nINNER LOOP:\n'
        header = '                      Kohn-Sham          SIC' \
                 '        Total             \n' \
                 '           time         energy:      energy:' \
                 '      energy:       Error:       G_max:'
        log(header0 + header)

    if outer_counter is not None:
        niter = outer_counter

    log('iter: %3d  %02d:%02d:%02d ' %
        (niter,
         t[3], t[4], t[5]
         ), end='')
    log('%11.6f  %11.6f  %11.6f  %11.1e  %11.1e' %
        (Hartree * eks,
         Hartree * esic,
         Hartree * (eks + esic),
         kappa,
         Hartree * g_max), end='')
    log(flush=True)


def g_max_norm(g_k, wfs):
    # get maximum of gradients
    n_kps = wfs.kd.nibzkpts

    max_norm = []
    for kpt in wfs.kpt_u:
        k = n_kps * kpt.s + kpt.q
        dim = g_k[k].shape[0]
        if dim == 0:
            max_norm.append(0.0)
        else:
            max_norm.append(np.max(np.absolute(g_k[k])))
    max_norm = np.max(np.asarray(max_norm))
    g_max = wfs.world.max_scalar(max_norm)

    return g_max
